import bluepyopt.ephys as ephys


class NrnFileMorphology_axon_fix(ephys.morphologies.NrnFileMorphology):

    def __init__(self,
                 morphology_path,
                 axon_length,
                 axon_diameter,
                 do_set_nseg=True,
                 comment='',
                 replace_axon_hoc=None,
                 axon_nseg_frequency=40,
                 nseg_frequency=40,
                 replace_axon_myelin_length=None,
                 replace_axon_myelin_diameter=None,
                 morph_modifiers=None,
                 morph_modifiers_hoc=None,
                 morph_modifiers_kwargs=None):

        super().__init__(morphology_path=morphology_path,
                         do_replace_axon=False,
                         do_set_nseg=do_set_nseg,
                         comment=comment,
                         replace_axon_hoc=replace_axon_hoc,
                         axon_nseg_frequency=axon_nseg_frequency*1e6,
                         nseg_frequency=nseg_frequency,
                         morph_modifiers=morph_modifiers,
                         morph_modifiers_hoc=morph_modifiers_hoc,
                         morph_modifiers_kwargs=morph_modifiers_kwargs)

        self.axon_diameter = axon_diameter
        self.axon_length = axon_length
        self.replace_axon_myelin_length = replace_axon_myelin_length
        self.replace_axon_myelin_diameter = replace_axon_myelin_diameter

    def instantiate(self, sim=None, icell=None):

        super().instantiate(sim=sim, icell=icell)

        self.replace_axon2(sim=sim, icell=icell)
        # Call the new replace_axon

    def replace_axon2(self, sim=None, icell=None):
        """Replace axon (length and diameter)"""

        for section in icell.axonal:
            sim.neuron.h.delete_section(sec=section)

        # Create new axon array
        sim.neuron.h.execute(f"create axon[{len(self.axon_length)}]", icell)

        assert len(self.axon_length) == len(self.axon_diameter)

        for index, (section, ax_len, ax_dia) in enumerate(zip(icell.axon, self.axon_length, self.axon_diameter)):
            section.L = ax_len * 1e6  # Convert to micrometers
            section.nseg = 1 + 2 * int(section.L / self.axon_nseg_frequency)
            section.diam = ax_dia * 1e6  # Convert to micrometers
            icell.axonal.append(sec=section)
            icell.all.append(sec=section)

        icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
        for index in range(1, len(self.axon_length)):
            icell.axon[index].connect(icell.axon[index-1], 1.0, 0.0)

        print(f"replace_axon2: {self.axon_length =}, {self.axon_diameter =}")

        if self.replace_axon_myelin_length is not None:
            print(f"Adding myelinated section to axon: {self.replace_axon_myelin_length =}, "
                  f"{self.replace_axon_myelin_diameter =}")

            for section in icell.myelinated:
                sim.neuron.h.delete_section(sec=section)

            sim.neuron.h.execute(f"create myelin[{len(self.replace_axon_myelin_length)}]", icell)

            assert len(self.replace_axon_myelin_length) == len(self.replace_axon_myelin_diameter)

            for index, (section, ax_len, ax_dia) in enumerate(zip(icell.myelin,
                                                                  self.replace_axon_myelin_length,
                                                                  self.replace_axon_myelin_diameter)):
                section.L = ax_len * 1e6  # Convert to micrometers
                section.nseg = 1 + 2 * int(section.L / self.axon_nseg_frequency)
                section.diam = ax_dia * 1e6  # Convert to micrometers
                icell.myelinated.append(sec=section)
                icell.all.append(sec=section)

            icell.myelin[0].connect(icell.axon[len(self.axon_length)-1], 1.0, 0.0)
            for index in range(1, len(self.replace_axon_myelin_length)):
                icell.myelin[index].connect(icell.myelin[index - 1], 1.0, 0.0)




# Please note, this is just for compability

from snudda.detect.prune import SnuddaPrune

print("SnuddaPrune is now available when doing import snudda.detect")
print("Please update to `from snudda.detect import SnuddaPrune`")



from snudda.place.place import SnuddaPlace
from snudda.place.rotation import SnuddaRotate
# from snudda.place.region_mesh_OLD import RegionMesh
from snudda.place.region_mesh_vedo import RegionMesh

from snudda.place.create_cube_mesh import create_cube_mesh
from snudda.place.create_slice_mesh import create_slice_mesh



import os
import re
import numpy as np
import h5py
import scipy
import scipy.optimize
import matplotlib.pyplot as plt
import matplotlib
import neuron
import json
import copy
import time

# !!! Add check that if the voltage is 0 or 5, then the trace is skipped entirely

from run_little_synapse_run import RunLittleSynapseRun

############################################################################

# JSON can not handle numpy arrays or numpy ints/floats, fix with a
# custom serialiser

class NumpyEncoder(json.JSONEncoder):
  def default(self, obj):
    print("NumpyEncoder: " + str(type(obj)))
      
    if isinstance(obj, np.integer):
      return int(obj)
    elif isinstance(obj, np.floating):
      return float(obj)
    elif isinstance(obj, np.ndarray):
      return obj.tolist()
    else:
      #return super(NumpyEncoder, self).default(obj)
      return json.JSONEncoder.default(self, obj)

############################################################################
          
class OptimiseSynapses(object):

  ############################################################################
  
  def __init__(self, fileName, synapseType="glut",loadCache=True,
               role="master",dView=None,verbose=True,logFileName=None,
               optMethod="swarm",prettyPlot=False):

    # Parallel execution role, "master" or "servant"
    self.role = role
    self.parallelSetupFlag = False # Set to True when servants are done setup
    self.dView = dView
    self.verbose = verbose
    self.logFileName = logFileName
    self.optMethod = optMethod
    self.nSmoothing = 200 # How many smoothing points do we use?
    self.simTime = 1.5
    
    self.debugParsFlag = False
    self.debugPars = []

    self.prettyPlot = prettyPlot
    
    print("Init optMethod = " + str(optMethod))
    
    if(self.logFileName is not None and len(self.logFileName) > 0):
      print("Log file: " + self.logFileName)
      self.logFile = open(self.logFileName,'w')
    else:
      self.logFile = None
      
    self.figResolution = 300

    self.fileName = fileName
    self.cacheFileName = str(self.fileName) + "-parameters.json"
    self.loadCache = loadCache
    self.synapseType = synapseType

    self.rsrSynapseModel = None
    self.rsrDeltaModel = None

    self.modelInfo = None

    if(loadCache):
      self.loadParameterCache()
    else:
      self.parameterCache = dict([])

    if(fileName is not None):
      self.hFile = h5py.File(fileName,"r")
      # self.hFile.swmr_mode = True # Allow multiple reads from h5py file
    
      self.writeLog("Loading " + str(fileName))

    if(self.role == "master"):
      self.setupParallel(dView=dView)


  ############################################################################
    
  def __delete__(self):

    if(self.hFile is not None):
      self.hFile.close()

    # Save the parameter cache before closing
    self.saveParameterCache()

    if(self.logFile is not None):
      self.logFile.close()
      self.logFile = None
    
  ############################################################################

  def saveParameterCache(self):

    if(self.role != "master"):
      self.writeLog("No servants are allowed to write output to json, ignoring call.")
      return
    
    self.writeLog("Saving parameters to cache file: " + str(self.cacheFileName))

    try:
      with open(self.cacheFileName,"w") as f:
        json.dump(self.parameterCache,f,indent=2,cls=NumpyEncoder)
        f.close()
    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)
      
      self.writeLog("Failed to save cache file ... " + str(self.cacheFileName))
      
  ############################################################################

  def loadParameterCache(self):

    if(os.path.exists(self.cacheFileName)):
      try:
        self.writeLog("Loading cache file " + str(self.cacheFileName))
        with open(self.cacheFileName,"r") as f:
          tmpDict = json.load(f)
          
          self.parameterCache = dict([])
          for k in tmpDict:
            self.parameterCache[int(k)] = tmpDict[k]

          f.close()
      except:
        import traceback
        tstr = traceback.format_exc()
        self.writeLog(tstr)
        
        self.writeLog("Unable to open " + str(self.cacheFileName))
        self.parameterCache = dict([])
    else:
      # No cache file to load, create empty dictionary
      self.parameterCache = dict([])

  ############################################################################

  def getSubDictionary(self,cellID):

    try:
      if(type(cellID) == list or type(cellID) == np.ndarray):
        retDict = dict([])

        for c in cellID:
          retDict[c] = copy.deepcopy(self.parameterCache[c])

        return retDict
    
      else:
        return self.parameterCache[cellID]
    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)
      import pdb
      pdb.set_trace()
      
  ############################################################################
  
  def addParameterCache(self,cellID,name,value):

    if(cellID not in self.parameterCache):
      self.parameterCache[int(cellID)] = dict([])
    
    self.parameterCache[int(cellID)][name] = value

  ############################################################################

  def getParameterCache(self,cellID,name):

    if(cellID in self.parameterCache and name in self.parameterCache[cellID]):
      return self.parameterCache[cellID][name]
    else:
      return None
    
  ############################################################################

  def getData(self,dataType,cellID=None):

    if(cellID is None):
      data = self.hFile[dataType].value.copy()
    else:
      data = self.hFile[dataType][:,cellID].copy()
      
    # data: 0 = no recording, 5 = recording but no response
      
    if("IGORWaveScaling" in self.hFile[dataType].attrs):
      tStep = self.hFile[dataType].attrs["IGORWaveScaling"][1,0]

      # 0 = no recording
      # 5 = has recording, no respons
      
      assert 0 < tStep and tStep < 1e-2, " Double check tStep = " + str(tStep)
      # Which variable contains the start time? Do not know, so assume it is 0
    
      nPoints = data.shape[0]
      t = 0 + tStep * np.arange(0,nPoints)

      return (data,t)

    else:
      
      return data

  ############################################################################

  def getCellType(self,cellID):

    # If the text is byte encoded, decode it first
    allCellTypes = [ x if type(x) != bytes else x.decode() \
                     for x in self.getData("w_cell").flatten()]

    return allCellTypes[cellID]
  
  ############################################################################
  
  def getCellID(self,dataType,cellType):

    # If the text is byte encoded, decode it first
    allCellTypes = [ x if type(x) != bytes else x.decode() \
                     for x in self.getData("w_cell").flatten()]
    
    
    # Just a check that our assumption on the ID is correct
    # tmp = self.getData("w_num").astype(int)
    # assert (tmp == np.arange(0,len(tmp)+1)).all(), "Numering mis-match!"
     
    cellIDs = [x for x,y in enumerate(allCellTypes) \
               if cellType.lower() in y.lower()]

    validIDs = self.getValidCellID(dataType)

    validCellIDs = np.array(list(set(cellIDs).intersection(validIDs)))
    
    return validCellIDs

  ############################################################################

  def getAllCellID(self):

    return np.arange(0,len(self.getData("w_cell").flatten()),dtype=int)

  ############################################################################

  def getValidCellID(self,dataType):

    allIdx = self.getAllCellID()

    helper = lambda x : self.checkValidData(dataType, x)
    validFlag = [x for x in map(helper,allIdx)]

    self.writeLog("Valid data: " + str(sum(validFlag)))
    
    return allIdx[validFlag]

  ############################################################################

  def getUserID(self,IDstring,dataType=None):

    IDlist = [int(x) for x in IDstring.split(",")]

    if(dataType is not None):
      validID = getValidCellID(dataType)
      validIDlist = [x for x in IDlist if x in validID]
      return validIDlist

    else:
      return IDlist
  
  ############################################################################

  # parDict is the parameters that are associated with cellID


  
  def plotData(self,dataType,cellID=None,params={},show=True,skipTime=0.05,
               prettyPlot=None):
      
    
    if(prettyPlot is None):
      prettyPlot = self.prettyPlot

    if(prettyPlot):
      matplotlib.rcParams.update({'font.size': 24})
    else:
      matplotlib.rcParams.update({'font.size': 5})
      
    (data,t) = self.getData(dataType,cellID)

    if(data is None):
      self.writeLog(dataType + " " + str(cellID) + ": Nothing to plot")
      return
    
    somaParams = self.getParameterCache(cellID,"soma")

    vPlot = None
    if(somaParams is not None):

      for p in somaParams:
        params[p] = somaParams[p]

      synapseParams = self.getParameterCache(cellID,"synapse")
      if(synapseParams is not None):
        
        for p in synapseParams:
          params[p] = synapseParams[p]
      
        # Get experimental data
        (inputResSteadyState,tauDelta,baselineDepol) \
          = self.getCellProperties(dataType,cellID)

        tStim = self.getStimTime(dataType,cellID)
    
        plotModel = RunLittleSynapseRun(stim_times=tStim,
                                        holding_voltage=baselineDepol,
                                        synapse_type=self.synapseType,
                                        params=params,
                                        time=self.simTime)
    
        (tPlot,vPlot,iPlot) = plotModel.run2(pars=params)

    tIdx = np.where(skipTime <= t)[0]
        
    plt.figure()

    plt.plot(t[tIdx]*1e3,data[tIdx]*1e3,'r-')
    if(vPlot is not None):
      t2Idx = np.where(skipTime <= tPlot)[0]
      plt.plot(tPlot[t2Idx]*1e3,vPlot[t2Idx]*1e3,'k-')
    # plt.title(dataType + " " + str(cellID))
    cellType = self.getCellType(cellID)
    if(not prettyPlot):
      titleStr = cellType + " " + str(cellID)

      if("nmda_ratio" in params):
        titleStr += "\nU=%.3g, tauR=%.3g, tauF=%.3g, tau=%.3g,\ncond=%.3g, nmda_ratio=%.3g" \
          % (params["U"],
             params["tauR"],
             params["tauF"],
             params["tau"],
             params["cond"],
             params["nmda_ratio"])
      else:
        titleStr += "\nU=%.3g, tauR=%.3g, tauF=%.3g, tau=%.3g,\ncond=%.3g" \
          % (params["U"],
             params["tauR"],
             params["tauF"],
             params["tau"],
             params["cond"])
        

      plt.title(titleStr)

    if(prettyPlot):
      # Draw scalebars
      vScaleX = 1200
      #vMax = np.max(vPlot[np.where(tPlot > 0.050)[0]])
      vBase = vPlot[-1]
      yScaleBar = vBase*1e3 + float(np.diff(plt.ylim()))/4
      vScaleY1 = yScaleBar+1
      vScaleY2 = yScaleBar
      tScaleY  = yScaleBar
      tScaleX1 = vScaleX
      tScaleX2 = vScaleX + 100
      
      plt.plot([vScaleX,vScaleX],[vScaleY1,vScaleY2],color="black")
      plt.plot([tScaleX1,tScaleX2],[tScaleY,tScaleY],color="black")

      plt.text(vScaleX-100, vScaleY2 + 0.20*float(np.diff(plt.ylim())),\
               ("%.0f" % (vScaleY1-vScaleY2)) + " mV",
                     rotation=90)
      plt.text(vScaleX,vScaleY2 - float(np.diff(plt.ylim()))/10,
               ("%.0f" % (tScaleX2 - tScaleX1) + " ms"))
      
      # Mark optogenetical stimulation
      yHeight = float(np.diff(plt.ylim()))/13
      
      tStim = self.getStimTime(dataType,cellID)*1e3
      yStimMarker1 = vPlot[-1]*1e3-1.5*yHeight
      yStimMarker2 = vPlot[-1]*1e3-2.5*yHeight
      for ts in tStim:
        plt.plot([ts,ts],[yStimMarker1,yStimMarker2],color="cyan")      

      plt.axis("off")

      #import pdb
      #pdb.set_trace()
      
      
    plt.xlabel("Time (ms)")
    plt.ylabel("Volt (mV)")
    
    if(not os.path.exists("figures/")):
      os.makedirs("figures/")
    
    baseName = os.path.splitext(os.path.basename(self.fileName))[0]
    figName = "figures/" + baseName + "-"  + dataType \
      + "-" + cellType + "-" + str(cellID) + ".pdf"
    plt.savefig(figName,dpi=self.figResolution)

    if(show):
      plt.ion()
      plt.show()
    else:
      plt.ioff()
      plt.close()
      
    
  ############################################################################

  def getCellProperties(self,dataType,cellID):

    data = self.getData("w_ecode",cellID)
    inputResSteadyState = data[32]*1e6
    tauDelta = data[40]*1e-3
    # baselineDepol = data[58]*1e-3
    baselineDepol = self.getExpBaseline(dataType,cellID)
    
    return (inputResSteadyState,tauDelta,baselineDepol)

  ############################################################################

  def extractInputResTau(self, t,v,curAmp,curStart,curEnd,baseStart,baseEnd):

    # We assume SI units
    tIdxBase = np.where(np.logical_and(baseStart < t, t < baseEnd))[0]
    vBase = np.mean([v[x] for x in tIdxBase])

    tIdxPeak = np.where(np.logical_and(curStart < t, t < curEnd))[0]
    vPeak = np.min(v[tIdxPeak])
    #vPeak = np.max([v[x] for x in tIdxPeak])

    assert np.abs(vPeak-vBase) > np.abs(np.max(v[tIdxPeak]) - vBase), \
      "The code assumes a hyperpolarising pulse, not a peak maximum"
    
    RM = (vPeak-vBase)/curAmp

    idxPostPulse = np.where(curEnd < t)[0]
    idxMaxPostPulse = idxPostPulse[np.argmax(v[idxPostPulse])]
    tMaxPostPulse = t[idxMaxPostPulse]
    
    tIdxDecay = np.where(np.logical_and(curEnd < t,t<tMaxPostPulse))[0]

    decayFunc = lambda x,a,b,c : a*np.exp(-x/b) + c

    tABfit = t[tIdxDecay] - t[tIdxDecay[0]]
    vABfit = v[tIdxDecay]

    p0d = [-0.005,0.01,-0.06]

    if(np.isnan(vABfit).any() or np.isinf(vABfit).any()):
      self.writeLog("We have inifinite or nan values in the voltage")
      import pdb
      pdb.set_trace()
    
    try:
      fitParams,pcov = scipy.optimize.curve_fit(decayFunc,tABfit,vABfit,
                                                p0=p0d)               
      tau = fitParams[1]

    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)

      plt.figure()
      plt.plot(tABfit,vABfit)
      plt.ion()
      plt.show()
      
      import pdb
      pdb.set_trace()        
      
      
    if(False):
      plt.figure()
      plt.plot(tABfit,vABfit,'-r')
      plt.plot(tABfit,decayFunc(tABfit,
                                   fitParams[0],
                                   fitParams[1],
                                   fitParams[2]))
      plt.xlabel("t")
      plt.ylabel("v")
      plt.title("Tau extraction")
      plt.ion()
      plt.show()                  

    # self.writeLog("RM = " + str(RM) + " tau = " + str(tau))

    # Return membrane resistance and tau
    return (RM,tau) 
    

  ############################################################################

  def setupModelDeltaPulse(self,dataType,cellID):

    # Experimental data from file (only baselineDepol needed at this point)
    (inputResSteadyState,tauDelta,baselineDepol) \
      = self.getCellProperties(dataType,cellID)

    # Initial guesses for parameters of soma
    somaDiam = 16e-6     # m
    somaGleak = 0.0003e4 # S/m2

    #(1.0 / tauDelta) / (np.pi*2*16*16) # !!! Check me!!

    params = { "somaDiameter" : somaDiam, "somaGleak" : somaGleak }
    self.rsrDeltaModel = RunLittleSynapseRun(stim_times=np.array([]),
                                             holding_voltage=baselineDepol,
                                             synapse_type=self.synapseType,
                                             params=params,
                                             time=self.simTime)
    # tau = R * C

    # This is the short hyperpolarising current pulse that we use to
    # estimate tau and input resistance
    # SI units
    self.curInj = -0.1e-9
    self.curStart = 0.5
    self.curDuration = 2e-3
    self.curEnd = self.curStart + self.curDuration
    self.baseStart = 0.2
    self.baseEnd = 0.3

    # Neuron uses natural units !
    self.rsrDeltaModel.IClamp2 = neuron.h.i_clamp(self.rsrDeltaModel.soma(0.5))
    self.rsrDeltaModel.IClamp2.amp = self.curInj*1e9
    self.rsrDeltaModel.IClamp2.delay = self.curStart*1e3
    self.rsrDeltaModel.IClamp2.dur = self.curDuration*1e3
    neuron.h.finitialize(self.rsrDeltaModel.holding_voltage * 1e3) # OLD : -70

  ############################################################################

  # x is somaDiameter, Gleak ...
  
  def optHelperFunction(self,x,inputResSteadyState,tauDelta,plotResults=False):

    try:
      somaDiameter,somaGleak = x
    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)

      import pdb
      pdb.set_trace()

    print("somaDiameter = " +str(somaDiameter))
    print("somaGleak = " + str(somaGleak))
    
    # Set soma diameter and Gleak (convert to natural units!)
    self.rsrDeltaModel.soma.L = somaDiameter*1e6
    self.rsrDeltaModel.soma.diam = somaDiameter*1e6
    self.rsrDeltaModel.soma.gl_hh = somaGleak *1e-4
    
    # We probably dont need to disable the short hyperpolarising pulse...
    
    # Update the holding current (use same holding voltage as before by default)
    self.rsrDeltaModel.update_holding_current()

    # Run the simulation    
    neuron.h.tstop = self.rsrDeltaModel.time *1e3# Must set tstop
    neuron.h.run()

    (RM,tau) =self.extractInputResTau(t=np.array(self.rsrDeltaModel.t_save) * 1e-3,
                                      v=np.array(self.rsrDeltaModel.v_save) * 1e-3,
                                      curAmp=self.curInj,
                                      curStart=self.curStart,
                                      curEnd=self.curEnd,
                                      baseStart=self.baseStart,
                                      baseEnd=self.baseEnd)

    # Return input resistance and tau    
    optError = np.abs((RM-inputResSteadyState)*(tauDelta-tau))

    if(plotResults):
      plt.figure()
      plt.plot(self.rsrDeltaModel.t_save,
               self.rsrDeltaModel.v_save)
      plt.xlabel("Time (ms)")
      plt.ylabel("Volt (mV)")
      plt.title("RM = " + str(RM) + ", tau = " + str(tau))
      plt.ion()
      plt.show()
    
    return optError
    
  ############################################################################
  
  def fitCellProperties(self,dataType,cellID,plotResults=True):

    params = self.getParameterCache(cellID,"soma")
    
    if(params is not None):
      return params
    
    import neuron

    # Experimental data from file
    (inputResSteadyState,tauDelta,baselineDepol) \
      = self.getCellProperties(dataType,cellID)

    # Setup neuron model
    self.setupModelDeltaPulse(dataType,cellID)

    p0 = [16e-6,0.0003e4]
    res = scipy.optimize.minimize(self.optHelperFunction,p0,
                                  args=(inputResSteadyState,tauDelta),
                                  bounds=[(5e-6,50e-6),(0.0003e3,0.0003e5)])

    somaDiameter = res.x[0]
    somaGleak = res.x[1]
    
    self.writeLog("Soma Diameter: " + str(somaDiameter))
    self.writeLog("Soma Gleak: " + str(somaGleak))

    somaDict = { "somaDiameter" : somaDiameter,
                 "somaGleak" : somaGleak }
    self.addParameterCache(cellID,"soma", somaDict)

    self.saveParameterCache()

    if(plotResults and len(res) == 2):
      self.optHelperFunction(res,
                             inputResSteadyState=inputResSteadyState,
                             tauDelta=tauDelta,
                             plotResults=True)
    
    # Remove old model
    self.rsrDeltaModel = None
 
    return somaDict
    
  ############################################################################

  
  def getPeakIdx(self, dataType,cellID,
                 firstSpike=0.1,delayToLast=None):

    pTime = self.getStimTime(dataType=dataType,
                              cellID=cellID,
                              firstSpike=firstSpike,
                              delayToLast=delayToLast)

    freq = 1.0/(pTime[1]-pTime[0])
    pWindow = 1.0/(2*freq)*np.ones(pTime.shape)
    pWindow[-1] *= 5
    
    peakInfo = self.findPeaksHelper(cellID=cellID,
                                    dataType=dataType,
                                    pTime=pTime,
                                    pWindow=pWindow)

    return peakInfo["peakIdx"]

  ############################################################################

  def getPeakIdx2(self,stimTime,time,volt):

    freq = 1.0/(stimTime[1]-stimTime[0])
    pWindow = 1.0/(2*freq)*np.ones(stimTime.shape)
    pWindow[-1] *= 5
    
    peakInfo = self.findPeaksHelper(pTime=stimTime,
                                    pWindow=pWindow,
                                    time=time,
                                    volt=volt)

    return peakInfo["peakIdx"]
   
   
  
  ############################################################################
    
  def getStimTime(self,dataType,cellID,
                  firstSpike=0.1,delayToLast=None):
    
    try:
      freq = float(re.findall(r'H\d+',dataType)[0][1:])
    except:
      self.writeLog("Unable to extract frequency from " + str(dataType))
      import pdb
      pdb.set_trace()

    if(delayToLast is None):
      if(freq == 20.0):
        delayToLast = 0.55
      else:
        delayToLast = 0.5
      
    pTime = 0.1 + np.arange(0,8)*1.0/freq
    pTime = np.append(pTime,pTime[-1] + delayToLast)

    return pTime
    
  ############################################################################

  # Find peaks within pStart[i] and pStart[i]+pWindow[i]
  # The value is not the amplitude of the peak, just the voltage at the peak
  
  def findPeaksHelper(self,pTime,pWindow,
                      cellID=None,dataType=None,
                      time=None,volt=None):

    if(cellID is not None):
      assert dataType is not None, "If you give cellID you need dataType"
      peakData = self.getParameterCache(cellID,"peaks")
      if(peakData is not None):
        return peakData
    
      (volt,time) = self.getData(dataType,cellID) 
    else:
      assert volt is not None and time is not None, \
        "Either use cellID to get time and volt of experimental data, or send time and volt explicitly"
      
    peakIdx = []
    peakTime = []
    peakVolt = []
    
    for pt,pw in zip(pTime,pWindow):
      tStart = pt
      tEnd = pt + pw

      tIdx = np.where(np.logical_and(tStart <= time,time <= tEnd))[0]

      if(self.synapseType == "glut"):
        pIdx = tIdx[np.argmax(volt[tIdx])]
      elif(self.synapseType == "gaba"):
        # We assume that neuron is more depolarised than -65, ie gaba is
        # also depolarising
        pIdx = tIdx[np.argmax(volt[tIdx])]
      else:
        self.writeLog("Unknown synapse type : " +str(self.synapseType))
        import pdb
        pdb.set_trace()
        
      peakIdx.append(int(pIdx))
      peakTime.append(time[pIdx])
      peakVolt.append(volt[pIdx])
      
    # Save to cache -- obs peakVolt is NOT amplitude of peak, just volt

    peakDict = { "peakIdx" : np.array(peakIdx),
                 "peakTime" : np.array(peakTime),
                 "peakVolt" : np.array(peakVolt)} # NOT AMPLITUDE

    if(cellID is not None):
      self.addParameterCache(cellID,"peaks",peakDict)
                           
    return peakDict
  # (peakIdx,peakTime,peakVolt)

  ############################################################################

  def findTraceHeights(self,time,volt,peakIdx):

    decayFunc = lambda x,a,b,c : a*np.exp(-x/b) + c
    
    vBase = np.mean(volt[int(0.3*peakIdx[0]):int(0.8*peakIdx[0])])

    peakHeight = np.zeros((len(peakIdx,)))
    peakHeight[0] = volt[peakIdx[0]] - vBase

    decayFits = []

    
    for idxB in range(1,len(peakIdx)):

      if(peakHeight[0] > 0):
        if(idxB < len(peakIdx) -1):        
          p0d = [0.06,0.05,-0.074]
        else:
          p0d = [1e-5,100,-0.074]

          if(self.synapseType == "gaba"):
            p0d = [1e-8,10000,-0.0798]  
      else:
        # In some cases for GABA we had really fast decay back
        if(idxB < len(peakIdx) -1):        
          p0d = [-0.06,0.05,-0.0798]
        else:
          p0d = [-1e-5,1e5,-0.0798]

      idxA = idxB -1

      peakIdxA = peakIdx[idxB-1] # Prior peak
      peakIdxB = peakIdx[idxB] # Next peak

      if(idxB < len(peakIdx) -1):
        # Not the last spike
      
        idxStart = int(peakIdxA*0.9 + peakIdxB*0.1)
        idxEnd = int(peakIdxA*0.1 + peakIdxB*0.9)
      else:
        # Last spike, use only last half of decay trace
        idxStart = int(peakIdxA*0.5 + peakIdxB*0.5)
        idxEnd = int(peakIdxA*0.05 + peakIdxB*0.95)

      try:
        assert idxStart < idxEnd
      except:
        import traceback
        tstr = traceback.format_exc()
        self.writeLog(tstr)

        plt.figure()
        plt.plot(time,volt)
        plt.xlabel("Time (error plot)")
        plt.ylabel("Volt (error plot)")
        plt.ion()
        plt.show()
        plt.title("ERROR!!!")
        import pdb
        pdb.set_trace()
        
      tAB = time[idxStart:idxEnd]
      vAB = volt[idxStart:idxEnd]        

      tABfit = tAB - tAB[0]
      vABfit = vAB
     
      try:

        try:
          fitParams,pcov = scipy.optimize.curve_fit(decayFunc,tABfit,vABfit,
                                                    p0=p0d)
        except:
          import traceback
          tstr = traceback.format_exc()
          self.writeLog(tstr)
          
          self.writeLog("!!! Failed to converge, trying with smaller decay constant")
          p0d[1] *= 0.01
          fitParams,pcov = scipy.optimize.curve_fit(decayFunc,tABfit,vABfit,
                                                    p0=p0d)
          
        tB = time[peakIdxB] - tAB[0]
        vBaseB = decayFunc(tB,fitParams[0],fitParams[1],fitParams[2])
        
        peakHeight[idxB] = volt[peakIdxB] - vBaseB

        vFit = decayFunc(tAB-tAB[0],fitParams[0],fitParams[1],fitParams[2])
        decayFits.append((tAB,vFit))
        
      except:
        
        self.writeLog("Check that the threshold in the peak detection before is OK")
        #self.plot(name)
        import traceback
        tstr = traceback.format_exc()
        self.writeLog(tstr)

        if(True):
          plt.figure()
          plt.plot(tAB,vAB,'r')
          plt.title("Error in findTraceHeights")
          plt.xlabel("time")
          plt.ylabel("volt")
          #plt.plot(tAB,vFit,'k-')
          plt.ion()
          plt.show()

        import pdb
        pdb.set_trace()

    #import pdb
    #pdb.set_trace()
    
        
    return peakHeight.copy(), decayFits,vBase
        
  ############################################################################

  def setupModelSynapseFitting(self,dataType,cellID,params={}):

    # If the delta model existed, clear it
    #self.rsrDeltaModel = None
    
    tStim = self.getStimTime(dataType,cellID)
    
    somaParams = self.getParameterCache(cellID,"soma")
    assert somaParams is not None, "You must fit soma parameters first"

    # !!! Should we also try and load synapse parameters from cache here??
    for p in somaParams:
      params[p] = somaParams[p]

    # Get experimental data
    (inputResSteadyState,tauDelta,baselineDepol) \
      = self.getCellProperties(dataType,cellID)

    # !!! We need to get the baseline depolarisation in another way

    self.rsrSynapseModel = RunLittleSynapseRun(stim_times=tStim,
                                               holding_voltage=baselineDepol,
                                               synapse_type=self.synapseType,
                                               params=params,
                                               time=self.simTime)

  ############################################################################

  def neuronSynapseSwarmHelper(self,pars,tSpikes,peakHeight,
                               smoothExpTrace):

    if(self.debugParsFlag):
      self.debugPars.append(pars)
    
    res = np.zeros((pars.shape[0]))
    
    for idx,p in enumerate(pars):
      peakH,tSim,vSim = self._neuronSynapseSwarmHelper(p,tSpikes)

      # Calculating error in peak height
      hDiff = np.abs(peakH - peakHeight)
      hDiff[0] *= 3
      hDiff[-1] *= 3
      hError = np.sum(hDiff)/len(hDiff)

      # Calculate error in decay fit
      simTrace,simTime = self.smoothingTrace(vSim,self.nSmoothing,
                                             time=tSim,
                                             startTime=self.decayStartFit,
                                             endTime=self.decayEndFit)

      # We only want to use the bit of the trace after max
      idxMax = np.argmax(smoothExpTrace)

      # We divide by number of points in vector, to get the average deviation
      # then we multiply by 10000 to get an error comparable to the others
      decayError = np.sum((smoothExpTrace[idxMax:] \
                           - simTrace[idxMax:])**2) \
                           /(self.nSmoothing-idxMax+1) * 2000

      if(False):
        plt.figure()
        plt.plot(smoothExpTrace[idxMax:])
        plt.plot(simTrace[idxMax:])
        plt.ion()
        plt.show()
        import pdb
        pdb.set_trace()
      
      res[idx] = hError + decayError
      
    return res

  ############################################################################

  def smoothingTrace(self,originalTrace,nParts,time=None,
                     startTime=None,endTime=None):

    if(time is not None):
      tFlag = np.ones((len(originalTrace),),dtype=bool)
      
      if(endTime is not None):
        tFlag[np.where(time > endTime)[0]] = False

      if(startTime is not None):
         tFlag[np.where(time < startTime)[0]] = False
         
      #tIdx = np.where(tFlag)[0]
      trace = originalTrace[tFlag]
      t = time[tFlag]
    else:
      trace = originalTrace
      t = time
    
    N = int(np.round(len(trace)/nParts))
    
    smoothTrace = np.convolve(trace, np.ones((N,))/N, mode='valid')

    idx = np.linspace(0,len(smoothTrace)-1,num=nParts,dtype=int)
    
    return smoothTrace[idx], t[idx]
  
  ############################################################################
  
  def _neuronSynapseSwarmHelper(self,
                               pars,
                               tSpikes):

    if(self.synapseType == "glut"):
      U,tauR,tauF,tauRatio,cond,nmdaRatio = pars
      tau = tauR*tauRatio
      params = { "nmda_ratio" : nmdaRatio }
    elif(self.synapseType == "gaba"):
      U,tauR,tauF,tauRatio,cond = pars
      tau = tauR*tauRatio
      params = {}
    else:
      self.writeLog("Unknown synapse type: " + str(self.synapseType))
      import pdb
      pdb.set_trace()
    
    peakHeights,tSim,vSim = self.synapseModelNeuron(tSpikes,U,
                                                    tauR,tauF,cond,tau,
                                                    params=params,
                                                    returnTrace=True)
    
    return peakHeights,tSim,vSim
  
  ############################################################################
  
  def neuronSynapseHelper(self,
                          tSpike,U,tauR,tauF,
                          tauRatio=None,
                          cond = 1e-7, tau=None):

    assert tauRatio is None or tau is None, \
      "Only one of tau and tauRatio should be defined"
    
    if(tauRatio is not None):
      tau = tauR*tauRatio
    elif(tau is None):
      assert False, "tau or tauRatio must be specified"

    peakHeights = self.synapseModelNeuron(tSpike,U,tauR,tauF,cond,tau)
    
    return peakHeights

   ############################################################################

   # !!! OBS tauRatio is inparameter
   
  def neuronSynapseHelperGlut(self,tSpike, 
                              U, tauR, tauF, tauRatio, cond, nmdaRatio, 
                              smoothExpTrace8, smoothExpTrace9, expPeakHeight,
                              returnType="peaks"):

    if(self.debugParsFlag):
      self.debugPars.append([U, tauR, tauF, tauRatio, cond, nmdaRatio])
    
    params = { "nmda_ratio" : nmdaRatio }
    tau = tauR * tauRatio
    
    peakH,tSim,vSim = self.synapseModelNeuron(tSpike,U,
                                              tauR,tauF,cond,tau,
                                              params=params,
                                              returnTrace=True)
    
    
    # Calculate error in decay fit
    simTrace8,simTime8 = self.smoothingTrace(vSim,self.nSmoothing,
                                             time=tSim,
                                             startTime=self.decayStartFit8,
                                             endTime=self.decayEndFit8)

    simTrace9,simTime9 = self.smoothingTrace(vSim,self.nSmoothing,
                                             time=tSim,
                                             startTime=self.decayStartFit9,
                                             endTime=self.decayEndFit9)
    # We only want to use the bit of the trace after max
    idxMax8 = np.argmax(smoothExpTrace8)
    idxMax9 = np.argmax(smoothExpTrace9)    
    
    # Calculating error in peak height
    hDiff = np.abs(peakH - expPeakHeight)
    hDiff[0] *= 3
    hDiff[-2] *= 2    
    hDiff[-1] *= 3
    
    hError = np.sum(hDiff)/len(hDiff)

    decayError8 = np.sum((smoothExpTrace8[idxMax8:] \
                          - simTrace8[idxMax8:])**2) \
                         /(self.nSmoothing-idxMax8+1) * 10000

    decayError9 = np.sum((smoothExpTrace9[idxMax9:] \
                         - simTrace9[idxMax9:])**2) \
                         /(self.nSmoothing-idxMax9+1) * 10000

    fitError = hError + decayError8 + decayError9
      

    if(False):
      peakBase = vSim[-1]
      plt.figure()
      plt.plot(tSim,vSim,'k-')
      plt.plot(simTime8,simTrace8,'y--')
      plt.plot(simTime8,smoothExpTrace8,'r--')
      plt.plot(simTime9,simTrace9,'y--')
      plt.plot(simTime9,smoothExpTrace9,'r--')
      
      for tp,expH,modH in zip(tSpike,expPeakHeight,peakH):
        plt.plot([tp,tp],[peakBase,expH+peakBase],'r-',linewidth=3)
        plt.plot([tp,tp],[peakBase,modH+peakBase],'b-')
      plt.title("hE = %g, dE8 = %g, dE9 = %g" \
                % (hError,decayError8,decayError9))
        
      plt.ion()
      plt.show()
      
      import pdb
      pdb.set_trace()

    if(returnType == "peaks"):
      return peakH
    elif(returnType == "error"):
      return fitError
    elif(returnType == "full"):
      return fitError,peakH,tSim,vSim
    else:
      assert False, "Unknown return type: " + str(returnType)
    
   ############################################################################
  
  def synapseModelNeuron(self,tSpike,U,tauR,tauF,cond,tau,
                         params = {},
                         returnTrace=False):

    # self.writeLog("Running neuron model")
    
    assert self.rsrSynapseModel is not None, \
      "!!! Need to call setupModelSynapseFitting first"

    # Should we make a copy of params, to not destroy it? ;)
    params["U"]    = U
    params["tauR"] = tauR
    params["tauF"] = tauF
    params["cond"] = cond
    params["tau"]  = tau

    #self.writeLog("params=" + str(params))
    
    (tSim,vSim,iSim) = \
      self.rsrSynapseModel.run2(pars=params)

    if(tSim.shape != vSim.shape):
      self.writeLog("Shape are different, why?!")
      import pdb
      pdb.set_trace()            
    
    peakIdx = self.getPeakIdx2(time=tSim,volt=vSim,stimTime=tSpike)
    peakHeight,decayFits,vBase = self.findTraceHeights(tSim,vSim,peakIdx)

    if(returnTrace):
      return peakHeight,tSim,vSim
    else:
      return peakHeight
   
  ############################################################################

  def getExpBaseline(self,dataType,cellID,tBefore=0.08):

     (volt,time) = self.getData(dataType,cellID)

     idx = np.where(time < tBefore)[0]

     return np.mean(volt[idx])
  
  
  ############################################################################

  def getModelBounds(self,cellID):

    cellType = self.getCellType(cellID)

    if("FS" in cellType.upper()):
      # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR), nmda_ratio
      modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5,0.1],
                     [1.0,2,2,0.9999999,1e-1,1])

    elif("MS" in cellType.upper()):
      # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR), nmda_ratio
      modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5,0.3],
                     [1.0,2,2,0.9999999,1e-1,6])

    elif("CHAT" in cellType.upper()):
      # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR), nmda_ratio
      modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5,2],
                     [1.0,2,2,0.9999999,1e-1,8])

    elif("LTS" in cellType.upper()):
      # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR), nmda_ratio
      modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5,2],
                     [1.0,2,2,0.9999999,1e-1,8])
    else:
      self.writeLog("Unknown celltype in " + str(cellType))
      import pdb
      pdb.set_trace()

    return modelBounds
    
  
  ############################################################################
  
  def fitTrace(self,dataType,cellID,optMethod=None):
    
    if(optMethod is None):
      optMethod = self.optMethod
    
    (volt,time) = self.getData(dataType,cellID)

    # This should NOT be empty, it should have AMPA/NMDA specific parameters
    # that we do not optimise for...
    params = dict([])
    
    synapseModel = self.setupModelSynapseFitting(dataType=dataType,
                                                 cellID=cellID,
                                                 params=params)




    peakIdx = self.getPeakIdx(dataType,cellID)
    tSpikes = time[peakIdx]
    
    sigma = np.ones(len(peakIdx))
    sigma[-1] = 1./3

    stimTime =  self.getStimTime(dataType=dataType,
                                 cellID=cellID)
    
    peakHeight,decayFits,vBase = self.findTraceHeights(time,volt,peakIdx)

    # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR)
    modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5],[1.0,2,2,0.9999999,1e-1])
    # !!! bestRandom should also see these bounds...

    try:
      
      if(optMethod == "stupid"):

        assert False, "Use the swarm method!"
        # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR)
        modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5],[1.0,2,2,0.9999999,1e-1])

        # This one gives nmda_ratio as parameter also
        #modelBounds = self.getModelBounds(cellID)
        
        # Call best random, to find a good set of starting points
        p0 = self.best_random(synapse_model=synapseModel,
                              cellID=cellID,
                              tPeak = stimTime,
                              h_peak= peakHeight,
                              model_bounds=modelBounds)


#        fitParams,pcov = scipy.optimize.curve_fit(self.neuronSynapseHelper,
#                                                  stimTime,peakHeight,
#                                                  sigma=sigma,
#                                                  absolute_sigma=False,
#                                                  p0=p0,
#                                                  bounds=modelBounds)
        fitParams,pcov = scipy.optimize.minimize(self.neuron_synapse_helper,
                                                 stimTime,
                                                 x0=p0,
                                                 bounds=modelBounds)

        # tau < tauR, so we use tauRatio for optimisation
        fitParams[3] *= fitParams[1] # tau = tauR * tauRatio

        modelHeights = self.neuron_synapse_helper(stimTime,
                                                  u=fitParams[0],
                                                  tau_r=fitParams[1],
                                                  tau_f=fitParams[2],
                                                  tau=fitParams[3],
                                                  cond=fitParams[4])


        self.write_log("Parameters: U = %.3g, tauR = %.3g, tauF = %.3g, tau = %.3g, cond = %3.g" % tuple(fitParams))

      elif(optMethod=="sobol"):

        if(self.debugParsFlag):
          self.debugPars = []
        
        modelBounds = self.getModelBounds(cellID)

        self.decayStartFit8 = 0.45
        self.decayEndFit8   = 0.8

        self.decayStartFit9 = 1.0
        self.decayEndFit9   = 1.3

        smoothExpVolt8,smoothExpTime8 \
          = self.smoothingTrace(volt,self.nSmoothing,
                                time=time,
                                startTime = self.decayStartFit8,
                                endTime = self.decayEndFit8)
        
        smoothExpVolt9,smoothExpTime9 \
          = self.smoothingTrace(volt,self.nSmoothing,
                                time=time,
                                startTime = self.decayStartFit9,
                                endTime = self.decayEndFit9)

        
        startPar = self.sobolScan(synapseModel=synapseModel,
                                  cellID=cellID,
                                  tPeak = stimTime,
                                  hPeak = peakHeight,
                                  modelBounds=modelBounds,
                                  smoothExpTrace8=smoothExpVolt8,
                                  smoothExpTrace9=smoothExpVolt9)

        func = lambda x : \
          self.neuronSynapseHelperGlut(tSpike=stimTime,
                                       U=x[0],
                                       tauR=x[1],
                                       tauF=x[2],
                                       tauRatio=x[3],
                                       cond=x[4],
                                       nmdaRatio=x[5],
                                       smoothExpTrace8=smoothExpVolt8,
                                       smoothExpTrace9=smoothExpVolt9,
                                       expPeakHeight=peakHeight,
                                       returnType="error")

        mBounds = [x for x in zip(modelBounds[0],modelBounds[1])]
        res = scipy.optimize.minimize(func,
                                      x0=startPar,
                                      bounds=mBounds)

        fitParams = res.x
        
#        fitParams,pcov = scipy.optimize.curve_fit(func,
#                                                  stimTime,peakHeight,
#                                                  sigma=sigma,
#                                                  absolute_sigma=False,
#                                                  p0=startPar,
#                                                  bounds=modelBounds)

        modelError,modelHeight,tSim,vSim = \
          self.neuronSynapseHelperGlut(stimTime,
                                       U=fitParams[0],
                                       tauR=fitParams[1],
                                       tauF=fitParams[2],
                                       tauRatio=fitParams[3],
                                       cond=fitParams[4],
                                       nmdaRatio=fitParams[5],
                                       smoothExpTrace8=smoothExpVolt8,
                                       smoothExpTrace9=smoothExpVolt9,
                                       expPeakHeight=peakHeight,
                                       returnType="full")
          

        # tau < tauR, so we use tauRatio for optimisation
        fitParams[3] *= fitParams[1] # tau = tauR * tauRatio

        self.writeLog("Parameters: U = %.3g, tauR = %.3g, tauF = %.3g, tau = %.3g, cond = %.3g, nmdaRatio = %.3g" % tuple(fitParams))
        self.writeLog("Model error: %g" % modelError)
   
      elif(optMethod=="swarm"):

        if(self.debugParsFlag):
          self.debugPars = []

        # Increase upper cond limit to 1? -- M1LH 11 might need it
          
        # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR), nmda_ratio
        #modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5,0.3],
        #               [1.0,2,2,0.9999999,1e-1,4])

        modelBounds = self.getModelBounds(cellID)

        self.decayStartFit = 1.0
        self.decayEndFit = 1.2
         
        smoothVolt,smoothTime = self.smoothingTrace(volt,self.nSmoothing,
                                                 time=time,
                                                 startTime = self.decayStartFit,
                                                 endTime = self.decayEndFit)

        import pyswarms as ps
        from pyswarms.utils.functions import single_obj as fx

        options = {"c1":0.5, "c2":0.3, "w":0.9} # default
        
        # Increased from 200 to 300, to 400
        if(self.synapseType == "glut"):
          nPart = 400
        else:
          nPart = 50
          
        optimizer = ps.single.GlobalBestPSO(n_particles=nPart,
                                            dimensions=len(modelBounds[0]),
                                            options=options,
                                            bounds=modelBounds)

        # Set iter to 50, lowered to 10
        cost, fitParams = optimizer.optimize(self.neuronSynapseSwarmHelper,
                                             iters=10,
                                             tSpikes = stimTime,
                                             peakHeight = peakHeight,
                                             smoothExpTrace=smoothVolt)

        modelHeight,tSim,vSim = \
          self._neuronSynapseSwarmHelper(fitParams,stimTime)
        
        # tau < tauR, so we use tauRatio for optimisation
        fitParams[3] *= fitParams[1] # tau = tauR * tauRatio

        if(len(fitParams) == 6):
          self.writeLog("Parameters: U = %.3g, tauR = %.3g, tauF = %.3g, tau = %.3g, cond = %3.g, nmda_ratio = %.3g" % tuple(fitParams))
        elif(len(fitParams) == 5):
          self.writeLog("Parameters: U = %.3g, tauR = %.3g, tauF = %.3g, tau = %.3g, cond = %3.g" % tuple(fitParams))
        else:
          self.writeLog("Unknown parameter format: " + str(fitParams))
          import pdb
          pdb.set_trace()
      else:
        self.writeLog("NO METHOD SELECTED!!!")
        exit(-1)
        
      self.writeLog("peakHeight = " + str(peakHeight))
      self.writeLog("modelHeight = " + str(modelHeight))
            
    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)
      import pdb
      pdb.set_trace()


    if(len(fitParams) == 6):
      self.addParameterCache(cellID,"synapse", \
                             { "U" : fitParams[0],
                               "tauR" : fitParams[1],
                               "tauF" : fitParams[2],
                               "tau"  : fitParams[3],
                               "cond" : fitParams[4],
                               "nmda_ratio" : fitParams[5] })
    elif(len(fitParams) == 5):
      self.addParameterCache(cellID,"synapse", \
                             { "U" : fitParams[0],
                               "tauR" : fitParams[1],
                               "tauF" : fitParams[2],
                               "tau"  : fitParams[3],
                               "cond" : fitParams[4] })
    else:
      print("Unknown fitParams format, unable to save paramters = " \
            +str(fitParams))
      import pdb
      pdb.set_trace()
      

    self.addParameterCache(cellID,"meta", \
                           { "type" : self.getCellType(cellID),
                             "experiment" : dataType,
                             "datafile" : self.fileName })

    if(self.debugParsFlag):
      self.plotDebugPars()
    
    # Deallocate Neuron model
    self.rsrSynapseModel = None
    
  ############################################################################

  def sobolScan(self,synapseModel,cellID,
                tPeak,hPeak,
                modelBounds,
                smoothExpTrace8, smoothExpTrace9,
                nTrials=100,loadParamsFlag=False):

    assert self.synapseType == "glut", \
      "GABA synapse not supported yet in new version"
    
    import chaospy
    distribution = chaospy.J(chaospy.Uniform(modelBounds[0][0],
                                             modelBounds[1][0]),
                             chaospy.Uniform(modelBounds[0][1],
                                             modelBounds[1][1]),
                             chaospy.Uniform(modelBounds[0][2],
                                             modelBounds[1][2]),
                             chaospy.Uniform(modelBounds[0][3],
                                             modelBounds[1][3]),
                             chaospy.Uniform(modelBounds[0][4],
                                             modelBounds[1][4]),
                             chaospy.Uniform(modelBounds[0][5],
                                             modelBounds[1][5]))
    
    USobol,tauRSobol,tauFSobol,tauRatioSobol,condSobol,nmdaRatioSobol \
      = distribution.sample(nTrials, rule="sobol")

    
    # tauSobol = np.multiply(tauRatioSobol,tauRSobol)

    minPars = None
    minError = np.inf
      
    if(loadParamsFlag):
      # If we should load params then do so first
      minPars = self.getParameterCache(cellID,"synapse")
            
      if(minPars is not None):
        minError = self.neuronSynapseHelperGlut(tPeak,
                                                U=minPars[0],
                                                tauR=minPars[1],
                                                tauF=minPars[2],
                                                tauRatio=minPars[3]/minPars[1],
                                                cond=minPars[4],
                                                nmdaRatio=minPars[5],
                                                smoothExpTrace8=smoothExpTrace8,
                                                smoothExpTrace9=smoothExpTrace9,
                                                expPeakHeight=hPeak,
                                                returnType="error")

    idx = 0
        
    # This for loop should be parallelised... please...
    for U,tauR,tauF,tauRatio,cond,nmdaRatio \
        in zip(USobol, tauRSobol, tauFSobol, tauRatioSobol, \
               condSobol, nmdaRatioSobol):

      idx += 1
      if(idx % 100 == 0):
        self.writeLog("%d / %d : minError = %g" % (idx, len(USobol),minError))
        self.writeLog(str(minPar))
      
      error = self.neuronSynapseHelperGlut(tPeak,U,tauR,tauF,tauRatio,
                                           cond,nmdaRatio,
                                           smoothExpTrace8=smoothExpTrace8,
                                           smoothExpTrace9=smoothExpTrace9,
                                           expPeakHeight=hPeak,
                                           returnType="error")
      try:      
        if(error < minError):
          minError = error
          minPar = np.array([U,tauR,tauF,tauRatio,cond,nmdaRatio])

      except:
        import traceback
        tstr = traceback.format_exc()
        self.writeLog(tstr)

        import pdb
        pdb.set_trace()
        
        # For big runs we do no want to give up. Let's try again...
        continue

    return minPar
      
  ############################################################################
  
  def bestRandom(self,synapseModel,cellID,
                 tPeak,hPeak,
                 modelBounds,
                 nTrials=5000,loadParamsFlag=False):

    assert nTrials >= 1, "nTrials should be a positive integer"

    minError = np.inf
    minPar = None

    for idx in range(0,nTrials):
      if(idx % 100 == 0):
        self.writeLog("Pre-trial : " + str(idx) + "/" + str(nTrials))

      if(idx == 0 and loadParamsFlag):
        # If we should load params then do so first
        pars = self.getParameterCache(cellID,"synapse")
      else:
        pars = None
        
      if(pars is not None):
        U    = pars["U"]
        tauR = pars["tauR"]
        tauF = pars["tauF"]
        tau  = pars ["tau"]
        cond = pars["cond"]
      else:
        U = np.random.uniform(modelBounds[0][0],modelBounds[1][0])
        tauR = np.random.uniform(modelBounds[0][1],modelBounds[1][1])
        tauF = np.random.uniform(modelBounds[0][2],modelBounds[1][2])
        tau = tauR * np.random.uniform(modelBounds[0][3],modelBounds[1][3])
        cond = np.random.uniform(modelBounds[0][4],modelBounds[1][4])

      try:
        peakHeights = self.synapseModelNeuron(tPeak,U,tauR,tauF,tau,cond)
      
        error = np.abs(peakHeights - hPeak)
        error[0] *= 3
        error[1] *= 2
        error[-1] *= 3
        error = np.sum(error)
      
        if(error < minError):
          minError = error
          minPar = np.array([U,tauR,tauF,tau/tauR,cond])

      except:
        import traceback
        tstr = traceback.format_exc()
        self.writeLog(tstr)
        
        # For big runs we do no want to give up. Let's try again...
        continue

        
    return minPar

  ############################################################################

  def optimiseCell(self,dataType,cellID):

    try:
      cellID = int(cellID)

      self.writeLog("Optimising " + str(dataType) + " ID = " + str(cellID))
    
      self.fitCellProperties(dataType,cellID)
      self.fitTrace(dataType,cellID)
      self.plotData(dataType,cellID,show=False)

      # We need to extract the dictionary associated with cellID and
      # return the result so that the main function can plug it into the
      # central parameter cache -- OR DO WE?

      return self.getSubDictionary(cellID)
    
    except:
      
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)
      self.addParameterCache(cellID,"error",tstr)
      

  ############################################################################

  def parallelOptimiseCells(self,dataType,cellIDlist=None):

    if(self.role == "master"):

      if(cellIDlist is None):
        cellIDlist = self.getValidCellID(dataType)

      self.printCellProperties(dataType,cellIDlist)
        
      if(self.dView is None):
        self.writeLog("We are in serial mode, use serial code..")
        for c in cellIDlist:
          self.optimiseCell(dataType,c)

        # Save parameters to file
        self.saveParameterCache()
        return

      self.writeLog("Optimising in parallel")
      
      self.setupParallel(self.dView)
      self.dView.scatter("cellIDlist",cellIDlist,block=True)
      self.dView.push({"dataType" : dataType},block=True)
      
      try:

        cmdStr = "res = ly.parallelOptimiseCells(dataType=dataType,cellIDlist=cellIDlist)"
        self.writeLog("Starting execution")
        self.dView.execute(cmdStr,block=True)
        self.writeLog("Execution finished, gathering results")
        #res = self.dView.gather("res",block=True)
        res = self.dView["res"]
      
        # res now contains a list of dictionaries, each dictionary from one worker, we need
        # to merge these
        self.writeLog("Results gathered, merging dictionary")
        for m in res:
          #self.parameterCache.update(m)
          for k in m:
            self.parameterCache[int(k)] = m[k]
          
        self.saveParameterCache()

        self.writeLog("Done.")
      except:
        import traceback
        tstr = traceback.format_exc()
        self.writeLog(tstr)
        import pdb
        pdb.set_trace()
        
    else:
      try:
        # Servant runs here
        for cellID in cellIDlist:
          self.optimiseCell(dataType,cellID)

        return self.getSubDictionary(cellIDlist)
      
      except:
        import traceback
        tstr = traceback.format_exc()
        self.writeLog(tstr)
        import pdb
        pdb.set_trace()
     
  ############################################################################

  def printCellProperties(self,dataType,cellID):

    inputResList = []
    tauDeltaList = []
    baselineDepolList = []
    
    for cid in cellID:

      inputResSteadyState,tauDelta,baselineDepol \
        = self.getCellProperties(dataType,cid)

      inputResList.append(inputResSteadyState)
      tauDeltaList.append(tauDelta)
      baselineDepolList.append(baselineDepol)

    inputRes = np.array(inputResList)
    tauDelta = np.array(tauDeltaList)
    baselineDepol = np.array(baselineDepolList)
      
    plt.figure()
    plt.scatter(inputRes*1e-6,tauDelta*1e3)
    plt.xlabel("Input resistance (Mohm)")
    plt.ylabel("Tau (ms)")
    plt.ion()
    plt.show()

    plt.figure()
    plt.scatter(inputRes*1e-6,baselineDepol*1e3)
    plt.xlabel("Input resistance (Mohm)")
    plt.ylabel("Baseline depolarisation (mV)")
    plt.ion()
    plt.show()

    
  
  ############################################################################
  
  def checkValidData(self,dataType,cellID):

    data = self.hFile[dataType][:,cellID]
    goodDataFlag = True
    
    if((data == 0).all()):
      self.writeLog("No recording for " + str(dataType) + " ID = " + str(cellID))
      goodDataFlag = False
      
    if((data == 5).all()):
      self.writeLog("No response for " + str(dataType) + " ID = " + str(cellID))
      goodDataFlag = False

    (inputResSteadyState,tauDelta,baselineDepol) = \
      self.getCellProperties(dataType,cellID)

    cellType = self.getCellType(cellID)
    
    if("LTS" in cellType.upper()):
      inputResMaxLimit = 1000e6
    else:
      inputResMaxLimit = 500e6
    
    if(np.isnan(inputResSteadyState) or np.isnan(tauDelta) \
       or inputResSteadyState > inputResMaxLimit
       or inputResSteadyState < 5e6
       or tauDelta > 5):
      self.writeLog("Bad values for input resistance and/or tauDelta for " \
                    + str(dataType) + " ID = " +str(cellID) \
                    + "(" + str(self.getCellType(cellID)) + ")")
      goodDataFlag = False
      
    return goodDataFlag
      
  ############################################################################
  
  def listDataTypes(self):

    return [x for x in self.hFile]

  ############################################################################

  def setupParallel(self,dView=None):

    assert self.role == "master", "Only master should call setupParallel"

    if(dView is None):
      self.writeLog("No dView, no parallel")
      return
    else:
      self.dView = dView

    if(self.parallelSetupFlag):
      # Already setup servants
      return
    
    with self.dView.sync_imports():
      from run_little_synapse_run import RunLittleSynapseRun
      from OptimiseSynapses import NumpyEncoder
      from OptimiseSynapses import OptimiseSynapses

    self.writeLog("Setting up workers: " \
          + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))

    # Create unique log file names for the workers
    if(self.logFileName is not None):
      engineLogFile = [self.logFileName + "-" \
                       + str(x) for x in range(0,len(self.dView))]
    else:
      engineLogFile = [[] for x in range(0,len(self.dView))]

    nWorkers = len(self.dView)
    self.dView.scatter("engineLogFile",engineLogFile) 
    
    self.dView.push({"fileName": self.fileName,
                     "synapseType" : self.synapseType,
                     "loadCache" : self.loadCache,
                     "role" : "servant"})

    cmdStr = "ly = OptimiseSynapses(fileName=fileName, synapseType=synapseType,loadCache=loadCache,role=role,logFileName=engineLogFile[0])"
    self.dView.execute(cmdStr,block=True)
    self.parallelSetupFlag = True

  ############################################################################

  def writeLog(self,text,flush=True): # Change flush to False in future, debug
    if(self.logFile is not None):
      self.logFile.write(text + "\n")
      
      if(self.verbose):
        print(text)
        
      if(flush):
        self.logFile.flush()
    else:
      if(self.verbose):
        print(text)  

  ############################################################################

  def plotDebugPars(self):

    try:
      nIter = len(self.debugPars)
      nPoints = self.debugPars[0].shape[0]
      nPars = self.debugPars[0].shape[1]

      assert nPars == 6, "Should be six parameters"

      Uall      = np.zeros((nPoints,nIter))
      tauR      = np.zeros((nPoints,nIter))
      tauF      = np.zeros((nPoints,nIter))
      tauRatio  = np.zeros((nPoints,nIter))
      cond      = np.zeros((nPoints,nIter))
      nmdaRatio = np.zeros((nPoints,nIter))

      for ctr,par in enumerate(self.debugPars):
        Uall[:,ctr]      = par[:,0]
        tauR[:,ctr]      = par[:,1]
        tauF[:,ctr]      = par[:,2]
        tauRatio[:,ctr]  = par[:,3]
        cond[:,ctr]      = par[:,4]
        nmdaRatio[:,ctr] = par[:,5]
      
      
      plt.figure()
      plt.plot(Uall,cond,'-')
      plt.xlabel("U")
      plt.ylabel("cond")

      plt.figure()
      plt.plot(tauR,tauF,'-')
      plt.xlabel("tauR")
      plt.ylabel("tauF")

      plt.figure()
      plt.plot(tauRatio,nmdaRatio)
      plt.xlabel("tauRatio")
      plt.ylabel("nmdaRatio")

      plt.figure()
      plt.plot(Uall)
      plt.xlabel("Uall")

      plt.figure()
      plt.plot(nmdaRatio)
      plt.xlabel("nmdaRatio")
      
      plt.ion()
      plt.show()
      
      import pdb
      pdb.set_trace()

    except:
      import traceback
      tstr = traceback.format_exc()
      self.writeLog(tstr)

      import pdb
      pdb.set_trace()
      
  ############################################################################
      
if __name__ == "__main__":

  import argparse
  parser = argparse.ArgumentParser(description="Extract synaptic parameters from electrophysiological recordings")
  parser.add_argument("file", help="HDF5 file to read in")
  parser.add_argument("--st", help="Synapse type (glut or gaba)",
                      choices=["glut","gaba"])
  parser.add_argument("--type", help="Cell types",
                      choices=["MSN","FS","LTS","CHAT"])
  parser.add_argument("--id",
                      help="Cell ID, comma separated no spaces, eg. 1,2,3,7")
  parser.add_argument("--optMethod",
                      help="Optimisation method",
                      choices=["sobol","stupid","swarm"],
                      default="sobol")
  parser.add_argument("--plot",action="store_true",
                      help="plotting previous optimised model")
  parser.add_argument("--prettyplot",action="store_true",
                      help="plotting traces for article")
  
  args = parser.parse_args()

  optMethod = args.optMethod
  
  print("Reading file : " + args.file)
  print("Synapse type : " + args.st)
  print("Optimisation method : " + optMethod)

  print("IPYTHON_PROFILE = " + str(os.getenv('IPYTHON_PROFILE')))
  
  if(os.getenv('IPYTHON_PROFILE') is not None \
     or os.getenv('SLURMID') is not None):
    from ipyparallel import Client
    rc = Client(profile=os.getenv('IPYTHON_PROFILE'),
                debug=False)
    print('Client IDs: ' + str(rc.ids))

    # http://davidmasad.com/blog/simulation-with-ipyparallel/
    # http://people.duke.edu/~ccc14/sta-663-2016/19C_IPyParallel.html
    dView = rc.direct_view(targets='all') # rc[:] # Direct view into clients
    lbView = rc.load_balanced_view(targets='all')
  else:
    dView = None

  
  logFileName = "logs/" + os.path.basename(args.file) + "-log.txt"
  if(not os.path.exists("logs/")):
    os.makedirs("logs/")
     
  
  
  # "DATA/Yvonne2019/M1RH_Analysis_190925.h5"
  ly = OptimiseSynapses(args.file,synapseType=args.st,dView=dView,role="master",
                        logFileName=logFileName,optMethod=optMethod)
  
  if(args.plot or args.prettyplot):

    if(args.prettyplot):
      prettyPlotFlag = True
    else:
      prettyPlotFlag = False
    
    assert args.id is not None, "You need to specify which trace(s) to plot"
    for id in ly.getUserID(args.id):
      print("Only plotting data for ID " + str(int(id)))

      ly.plotData("GBZ_CC_H20",int(id),show=True,prettyPlot=prettyPlotFlag)
    exit(0)
    
  if(args.id is not None):
    # ID has priority over type
    userID = ly.getUserID(args.id)
    print("Optimising only " + str(userID))
    ly.parallelOptimiseCells("GBZ_CC_H20", userID)

  elif(args.type):
    print("Optimising only " + args.type + " in " + "GBZ_CC_H20")
    ly.parallelOptimiseCells("GBZ_CC_H20",
                             ly.getCellID("GBZ_CC_H20",args.type))
  else:
    ly.parallelOptimiseCells("GBZ_CC_H20")

    
  # (d2,t2) = ly.getData("GBZ_CC_H20",21)
  # xx = ly.getCellID("MSN")



  if(False):
    optHelper = lambda x : ly.optimiseCell( "GBZ_CC_H20", x)

    for idx in ly.get_all_cell_id():
      try:
        if(not ly.checkValidData("GBZ_CC_H20", idx)):
          # Skip invalid trace
          continue
        ly.optimise_cell("GBZ_CC_H20", idx)
      except:
        import traceback
        tstr = traceback.format_exc()
        print(tstr)
        print("!!! Something went wrong. Skipping neuron " + str(idx))
  
    ly.save_parameter_cache()
  
  print("Did we get all the parameters in the file?")
  #import pdb
  #pdb.set_trace()



import os
import numpy as np
import time


from OptimiseSynapses import OptimiseSynapses


class OptimiseSynapsesPlanert(OptimiseSynapses):

  def __init__(self,traceList,dView=None,logFileName="logs/planert.txt",
               role="master"):

    self.traceList = traceList
    
    # Create object
    super().__init__(fileName=None,synapseType="gaba",loadCache=False,
                     dView=dView,logFileName=logFileName,role=role)

    self.cacheFileName = traceList + "-parameters.json"
    self.data = {}
    self.nameList = []
    
    # Next load the data
    self.loadVonly(traceList)


    
    # import pdb
    # pdb.set_trace()
    
  ############################################################################
    
  def loadVonly(self,traceList):


    print("Reading traceList: " + str(traceList))
    self.fileName = traceList
    self.traceList = None

    with open(traceList) as f:
      tt = f.readlines()

    dataPath = os.path.dirname(traceList)
      
    for row in tt:
      rowdata = row.split()

      tracefile = rowdata[0]
      dt = float(rowdata[1]) * 1e-6
      timeToFirst = float(rowdata[2]) * 1e-3
      freq = float(rowdata[3])
      nSpikes = int(rowdata[4])
      delayToLast = float(rowdata[5]) * 1e-3

      name = tracefile.replace(".txt","")
      self.nameList.append(name)
      
      tFile = dataPath + "/" + tracefile

      if(not os.path.isfile(tFile)):
        tFileAlt = tFile + "_amp.dat"
        if(os.path.isfile(tFileAlt)):
          # We do not have the file, but we have the extracted amplitudes
          dd = np.genfromtxt(tFileAlt)

          self.data[name] = dict([])
          self.data[name]["tPeak"] = dd[:,0]
          self.data[name]["peakHeight"] = dd[:,1]
          print("Loaded only peaks from " + tFileAlt)
          continue
        else:
          print("File not found: " + tFile)
          continue
      else:
        print("Loading " + tFile)
      
      volt = np.genfromtxt(tFile)
      
      self.data[name] = dict([])
      
      self.data[name]["nPoints"] = nSpikes #  +1 for control spike?
      self.data[name]["time"] = dt*np.arange(0,len(volt))
      self.data[name]["voltage"] = volt

      print("Loaded " + tracefile)
      
      # Find peaks...
      peakTrigger = [timeToFirst]
      for idx in range(1,nSpikes):
        peakTrigger.append(peakTrigger[-1] + 1.0/freq)

      peakTrigger.append(peakTrigger[-1] + delayToLast)

      pTimes = np.array(peakTrigger)

      pWindow = 1.0/(2*freq)*np.ones(pTimes.shape)
      pWindow[-1] *= 5

      self.setPeakInWindow(name,pTimes,pWindow)
      
    return freq      

  ############################################################################
  
  # Look for peaks within a short window (pWindow) after specific time points
  
  def setPeakInWindow(self,name,pTimes,pWindow):

    if(not "time" in self.data[name] or not "voltage" in self.data[name]):
      print("Time or voltage missing from " + str(name))
      return
    
    t = self.data[name]["time"]
    v = self.data[name]["voltage"]
    
    peakIdx = []
    peakTime = []
    peakValue = []
    
    for pt,pw in zip(pTimes,pWindow):
      tStart = pt
      tEnd = pt + pw

      tIdx = np.where(np.logical_and(tStart <= t,t <= tEnd))[0]
      pIdx = tIdx[np.argmax(v[tIdx])]

      peakIdx.append(pIdx)
      peakTime.append(t[pIdx])
      peakValue.append(v[pIdx])

    self.data[name]["peakIdx"] = np.array(peakIdx)
    self.data[name]["tPeak"] = np.array(peakTime)
    self.data[name]["vPeak"] = np.array(peakValue)

  ############################################################################

  def getData(self,dataType,cellID):

    # Ignore dataType, used in data format from Yvonne
    name = self.nameList[cellID]

    return self.data[name]["voltage"],self.data[name]["time"]

  ############################################################################

  def getModelBounds(self,cellID):

    # U, tauR, tauF, tauRatio, cond (obs, tau = tauRatio * tauR)
    modelBounds = ([1e-3,1e-4,1e-4,0, 1e-5],
                   [1.0,2,2,0.9999999,1e-1])

    return modelBounds

  ############################################################################

  # We do not know these properties for the neuron, so assume some values
  # that are not too far off for MS and FS
  
  def getCellProperties(self,dataType,cellID):

    inputResSteadyState = 60e6
    tauDelta = 5e-3

    v,t = self.getData(dataType,cellID)
    baselineDepol = np.mean(v[np.where(t < 0.05)][0])

    return (inputResSteadyState,tauDelta,baselineDepol)
  
  ############################################################################

  def getStimTime(self,dataType,cellID,firstSpike=None,delayToLast=None):

    # Ignoring all parameters but cell ID
    
    stimTime = np.array([0.1,0.15,0.2,0.25,0.3,0.35,0.4,0.45,1.0])
    
    return stimTime

  ############################################################################

  def getCellType(self,cellID):

    return self.nameList[cellID].split("_")[0]

  ############################################################################

  def setupParallel(self,dView=None):

    assert self.role == "master", "Only master should call setupParallel"

    if(dView is None):
      self.writeLog("No dView, no parallel")
      return
    else:
      self.dView = dView

    if(self.parallelSetupFlag):
      # Already setup servants
      return
    
    with self.dView.sync_imports():
      from run_little_synapse_run import RunLittleSynapseRun
      from OptimiseSynapses import NumpyEncoder
      from OptimiseSynapsesPlanert import OptimiseSynapsesPlanert

    self.writeLog("Setting up workers: " \
          + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))

    # Create unique log file names for the workers
    if(self.logFileName is not None):
      engineLogFile = [self.logFileName + "-" \
                       + str(x) for x in range(0,len(self.dView))]
    else:
      engineLogFile = [[] for x in range(0,len(self.dView))]

    nWorkers = len(self.dView)
    self.dView.scatter("engineLogFile",engineLogFile) 
    
    self.dView.push({"traceList": self.traceList,
                     "role" : "servant"})

    cmdStr = "ly = OptimiseSynapsesPlanert(traceList=traceList,role=role,logFileName=engineLogFile[0])"
    self.dView.execute(cmdStr,block=True)
    self.parallelSetupFlag = True

  
  ############################################################################
    
if __name__ == "__main__":

  assert False, "This code uses the old incomplete Planert data, for the article we instead resorted to surrogate data. See Planert2010.py and Planert2010part2.py"
  
  print("IPYTHON_PROFILE = " + str(os.getenv('IPYTHON_PROFILE')))
  
  if(os.getenv('IPYTHON_PROFILE') is not None \
     or os.getenv('SLURMID') is not None):
    from ipyparallel import Client
    rc = Client(profile=os.getenv('IPYTHON_PROFILE'),
                debug=False)
    print('Client IDs: ' + str(rc.ids))

    # http://davidmasad.com/blog/simulation-with-ipyparallel/
    # http://people.duke.edu/~ccc14/sta-663-2016/19C_IPyParallel.html
    dView = rc.direct_view(targets='all') # rc[:] # Direct view into clients
    lbView = rc.load_balanced_view(targets='all')
  else:
    dView = None  
  
  traceList = "DATA/Planert2010/d1d2conns/traces/trace_table.txt"
  logFile = "logs/Planert-log.txt"

  osp = OptimiseSynapsesPlanert(traceList,dView=dView,
                                logFileName=logFile)
  
  osp.parallel_optimise_cells("planert",
                              np.arange(0,len(osp.nameList)))

  ops.save_parameter_cache()
  
  # osp.optimiseCell("planert",0)

  #import pdb
  #pdb.set_trace()



import unittest
import os
import numpy as np
import multiprocessing

try:
    multiprocessing.set_start_method("spawn", force=True)
except RuntimeError:
    # Already set, fine
    pass

from snudda import Snudda
from snudda.utils import SnuddaLoadSimulation

# TODO: Write example that uses Anu's SBML Dopamine cascade
# https://www.ebi.ac.uk/biomodels/BIOMD0000000636#Files
# Skriv SBML --> vårt json format, converter
# https://modeldb.science/237653?tab=2&file=Beskow/speedy_reduced2.mod
#
# Use: https://github.com/sbmlteam/libsbml

def run_sim(network_path, output_file):
    print(f"Running simulation: {output_file}")

    import os
    print("CWD in run_sim:", os.getcwd())

    mech_dir = "../validation/mechanisms"  # Added the kirrxd and DASyn as symbolic links to mechanisms

    # os.system(f"nrnivmodl {mech_dir}")

    snudda = Snudda(network_path=network_path)
    # snudda.compile_mechanisms(mech_dir=mech_dir)

    from neuron import h, nrn

    so_path = os.path.join("aarch64", ".libs", "libnrnmech.so")
    h.nrn_load_dll(so_path)

    print("Mechanisms loaded:")
    print("kirrxd visible?", hasattr(h, "kirrxd"))

    sim = snudda.simulate(time=0, mech_dir=mech_dir, use_rxd_neuromodulation=True)

    n = sim.neurons[0]

    sim.add_rxd_concentration_recording(species="DA", neuron_id=0,
                                        region="soma_internal",
                                        sec_id=-1,
                                        sec_x=0.5)

    sim.add_rxd_concentration_recording(species="B", neuron_id=0,
                                        region="soma_internal",
                                        sec_id=-1,
                                        sec_x=0.5)

    sim.add_rxd_concentration_recording(species="PKA", neuron_id=0,
                                        region="soma_internal",
                                        sec_id=-1,
                                        sec_x=0.5)

    sim.add_density_mechanism_recording(density_mechanism="kirrxd",
                                        variable="modulation_factor",
                                        neuron_id=0,
                                        sec_id=-1,
                                        sec_x=0.5)

    sim.add_density_mechanism_recording(density_mechanism="kirrxd",
                                        variable="m",
                                        neuron_id=0,
                                        sec_id=-1,
                                        sec_x=0.5)

    sim.add_membrane_recording(variable="PKAi",
                               neuron_id=0,
                               sec_id=-1,
                               sec_x=0.5)

    # Add DA synapse
    # da_syn = h.DASyn(soma(0.5))
    mod_file = "DASyn"
    eval_str = f"sim.sim.neuron.h.{mod_file}"
    channel_module = eval(eval_str)

    da_syn = sim.get_external_input_synapse(channel_module=channel_module,
                                            section=sim.neurons[0].icell.soma[0],
                                            section_x=0.5)
    da_syn.tau = 1

    net_stim = sim.sim.neuron.h.NetStim()
    net_stim.number = 100
    net_stim.start = 300
    net_stim.interval = 5

    nc = sim.sim.neuron.h.NetCon(net_stim, da_syn)
    nc.weight[0] = 100_000_000.0  # units : molecules/ms

    sim.neurons[0].modulation.link_synapse(species_name="DA",
                                           region="soma_internal",
                                           synapse=da_syn,
                                           flux_variable="open")

    sim.run(t=1000)

    # output_file = os.path.join(self.network_path, "simulation", "output-2.hdf5")
    sim.record.set_new_output_file(output_file)
    sim.record.write()

    # Important we need to delete the old da_syn
    del da_syn
    del net_stim
    del nc

    sim.clear_neuron()


class NeuromodulationTestCase(unittest.TestCase):

    def setUp(self):

        test_path = "test_project2"
        if os.path.isdir(test_path):
            import shutil
            shutil.rmtree(test_path)
        os.mkdir(test_path)
        os.chdir(test_path)

        self.neuron_path = "../validation/dspn_neurons_rxd"
        self.network_path = "networks/network_rxd"

        self.snudda = Snudda(network_path=self.network_path)
        self.snudda.init_tiny(neuron_paths=self.neuron_path,
                              neuron_names="neuron",
                              number_of_neurons=[10],
                              random_seed=123456)
        self.snudda.create_network()

        # Check why file size is so large, and why it is so slow to generate!

        # mech_dir = "../validation/mechanisms_rxd"

        mech_path = os.path.join(os.path.dirname(__file__), "validation/mechanisms")
        os.system(f"nrnivmodl {mech_path}")


    def test_reaction(self):

        output_file = os.path.join(self.network_path, "simulation", "output-2.hdf5")

        p = multiprocessing.Process(target=run_sim, args=(self.network_path, output_file,))
        p.start()
        p.join()

        nd = SnuddaLoadSimulation(output_file)
        time = nd.get_time()
        data_a = nd.get_data("DA", 0)
        data_b = nd.get_data("B", 0)
        data_ab = nd.get_data("PKA", 0)

        data_kir_modulation_factor = nd.get_data("kirrxd.modulation_factor", 0)[0][0]
        data_kir_m = nd.get_data("kirrxd.m", 0)[0][0]
        data_voltage = nd.get_data("voltage", 0)[0][0]
        data_pka = nd.get_data("membrane.PKAi", 0)[0][0]

        self.assertTrue(np.max(np.abs(data_a[0][0][0] - 0)) < 1e-7)
        self.assertTrue(np.max(np.abs(data_b[0][0][0] - 0.7)) < 1e-7)
        self.assertTrue(np.max(np.abs(data_ab[0][0][0] - 0.1)) < 1e-7)

        # self.assertTrue(data_a[0][0][-1] < data_a[0][0][0])
        # self.assertTrue(data_b[0][0][-1] < data_b[0][0][0])
        # self.assertTrue(data_ab[0][0][-1] > data_ab[0][0][0])

        # Plot A, B, PKA activity
        da = data_a[0][0]
        db = data_b[0][0]
        dab = data_ab[0][0]

        self.plot_data(time, np.hstack([da, db, dab]), legend=["DA", "B", "PKA"],
                       ylabel="Concentration", filename="concentration.png")

        # Plot the voltage and KIR modulation
        self.plot_data(time, np.hstack([data_kir_modulation_factor,
                                        data_kir_m, data_pka]),
                       legend=["kir_mod_factor", "kir_m", "membrane.pka"],
                       filename="kir_activation.png")

        self.plot_data(time, data_voltage,
                       legend=["voltage"],
                       filename="voltage.png")

    def plot_data(self, time, data, legend, filename=None, ylabel=None):
        import matplotlib.pyplot as plt
        plt.figure()
        plt.plot(time, data, label=legend)
        plt.xlabel("Time (s)")
        plt.ylabel(ylabel)
        plt.legend()

        if filename is not None:
            plt.savefig(filename, dpi=300)

        plt.ion()
        plt.show()

    def plot_kir_data(self):
        pass

    def tearDown(self):
        os.chdir("..")

        return

        print("\n[DEBUG] Running tearDown() for", self.id())

        # Remember to clear old neuron, for next unit test!
        self.sim.clear_neuron()

        # Really trying to make sure noting remains in NEURON
        # --- Hard NEURON reset to avoid stale rxd or mod pointer references ---
        from neuron import h
        import gc

        # Clear all rxd species, regions, and reactions
        try:
            from neuron import rxd
            rxd.region._all_regions.clear()
            rxd.species._all_species.clear()
            rxd.reaction._all_reactions.clear()
            rxd.rxdmath._all_reactions.clear()
            rxd.node._all_nodes.clear()
            rxd.include_flux._all_include_fluxes.clear()
        except Exception:
            pass

        # Reset ParallelContext and clear gids
        try:
            pc = h.ParallelContext()
            pc.gid_clear()
            del pc
        except Exception:
            pass

        # Delete any remaining sections and point processes
        for sec in list(h.allsec()):
            h.delete_section(sec=sec)
        h('forall delete_section()')

        # Force NEURON to release internal handles
        h('objref nil')
        h('forall delete_section()')
        h('nrnpython("import gc; gc.collect()")')

        gc.collect()

        import pdb
        pdb.set_trace()


if __name__ == '__main__':
    unittest.main()



# This script visualises two neurons and the location of the synapses
# connecting them.

import bpy
import os
import mathutils
import numpy as np
from snudda.utils.load import SnuddaLoad
from snudda.utils.snudda_path import snudda_parse_path


class VisualiseNetworkOld(object):

    # You need to provide neuron
    def __init__(self, network_path, blender_save_file=None, blender_output_image=None,
                 network_json=None):

        self.network_path = network_path

        if network_json:
            self.network_json = network_json
            self.network_file = None
        else:
            self.network_json = None
            self.network_file = os.path.join(network_path, "network-synapses.hdf5")

        if blender_save_file:
            self.blender_save_file = blender_save_file
        else:
            self.blender_save_file = os.path.join(network_path, "visualise-network.blend")

        self.blender_output_image = blender_output_image

        self.neuron_cache = dict([])

        # Load the neuron positions
        if self.network_file:
            self.sl = SnuddaLoad(self.network_file)
            self.data = self.sl.data
        elif self.network_json:
            from snudda.utils.fake_load import FakeLoad
            self.sl = FakeLoad()
            self.sl.import_json(self.network_json)
            self.data = self.sl.data

    def visualise(self, neuron_id=None, blender_output_image=None, white_background=True,
                  show_synapses=True):

        if neuron_id:
            neurons = [self.data["neurons"][x] for x in neuron_id]
        else:
            neurons = self.data["neurons"]
            neuron_id = self.data["neuronID"]

        if blender_output_image:
            self.blender_output_image = blender_output_image

        origo = self.data["simulationOrigo"]
        voxel_size = self.data["voxelSize"]

        # Remove the start cube
        # bpy.ops.object.delete()
        VisualiseNetworkOld.clean_scene()

        bpy.data.scenes['Scene'].render.engine = 'CYCLES'
        world = bpy.data.worlds['World']
        world.use_nodes = True

        # changing these values does affect the render.
        bg = world.node_tree.nodes['Background']
        if white_background:  # Set to True for white background
            bg.inputs[0].default_value[:3] = (1.0, 1.0, 1.0)
            bg.inputs[1].default_value = 1.0
        else:
            bg.inputs[0].default_value[:3] = (0.0, 0.0, 0.0)
            bg.inputs[1].default_value = 0.0

        # Define materials
        mat_msd1 = bpy.data.materials.new("PKHG")
        mat_msd1.diffuse_color = (77. / 255, 151. / 255, 1.0)
        mat_msd2 = bpy.data.materials.new("PKHG")
        mat_msd2.diffuse_color = (67. / 255, 55. / 255, 181. / 255)
        mat_fs = bpy.data.materials.new("PKHG")
        mat_fs.diffuse_color = (6. / 255, 31. / 255, 85. / 255)
        mat_chin = bpy.data.materials.new("PKHG")
        mat_chin.diffuse_color = (252. / 255, 102. / 255, 0.0)
        mat_lts = bpy.data.materials.new("PKHG")
        mat_lts.diffuse_color = (150. / 255, 63. / 255, 212. / 255)
        mat_other = bpy.data.materials.new("PKHG")
        mat_other.diffuse_color = (0.4, 0.4, 0.4)

        mat_synapse = bpy.data.materials.new("PKHG")

        material_lookup = { "dspn": mat_msd1,
                            "ispn": mat_msd2,
                            "fsn": mat_fs,
                            "fs": mat_fs,
                            "chin": mat_chin,
                            "lts": mat_lts,
                            "synapse": mat_synapse,
                            "other": mat_other}

        if white_background:
            mat_synapse.diffuse_color = (0.8, 0.0, 0.0)
        else:
            mat_synapse.diffuse_color = (1.0, 1.0, 0.9)

        # matSynapse.use_transparency = True
        mat_synapse.use_nodes = True

        if not white_background:
            emission_strength = 5.0

            # Make synapses glow
            emission = mat_synapse.node_tree.nodes.new('ShaderNodeEmission')
            emission.inputs['Strength'].default_value = emission_strength

            material_output = mat_synapse.node_tree.nodes.get('Material Output')
            mat_synapse.node_tree.links.new(material_output.inputs[0], emission.outputs[0])

        for neuron in neurons:

            e_rot = mathutils.Matrix(neuron["rotation"].reshape(3, 3)).to_euler()

            if neuron["name"] in self.neuron_cache:
                # If we already have the object in memory, copy it.
                obj = self.neuron_cache[neuron["name"]].copy()

                if self.neuron_cache[neuron["name"]].data:
                    obj.data = self.neuron_cache[neuron["name"]].data.copy()

                VisualiseNetworkOld.copy_children(self.neuron_cache[neuron["name"]], obj)
                obj.animation_data_clear()

                # Will return None if there is no obj named CUBe
                obj.name = f"{neuron['name']}-{neuron['neuronID']}"
                VisualiseNetworkOld.link_object(obj)
            else:
                bpy.ops.import_mesh.swc(filepath=snudda_parse_path(neuron["morphology"]))
                obj = bpy.context.selected_objects[0]
                obj.name = f"{neuron['name']}-{neuron['neuronID']}"

                self.neuron_cache[neuron["name"]] = obj

            obj.rotation_euler = e_rot
            print(f"Setting neuron {neuron['neuronID']} ({neuron['name']}) position: {neuron['position'] * 1e3}")
            obj.location = neuron["position"] * 1e3

            n_type = neuron["type"].lower()
            if n_type in material_lookup:
                mat = material_lookup[n_type]
            else:
                mat = material_lookup["other"]

            for ch in obj.children:
                ch.active_material = mat

            obj.select = False

        if show_synapses:
            print("Adding synapses...")

            # Draw the synapses
            n_synapses = 0

            for ob in bpy.context.selected_objects:
                ob.select = False

            synapse_obj = None

            for vis_pre_id in neuron_id:
                for vis_post_id in neuron_id:
                    synapses, synapse_coords = self.sl.find_synapses(pre_id=vis_pre_id, post_id=vis_post_id)

                    if synapses is None:
                        # No synapses between pair
                        continue

                    for syn in synapses:
                        pre_id = syn[0]
                        post_id = syn[1]

                        assert pre_id == vis_pre_id and post_id == vis_post_id  # Just sanity check, should be true

                        # Draw this neuron (the SWC import scales from micrometers to mm), the
                        # positions in the simulation are in meters, need to scale it to mm for
                        # blender to have same units.
                        x = (origo[0] + voxel_size * syn[2]) * 1e3
                        y = (origo[1] + voxel_size * syn[3]) * 1e3
                        z = (origo[2] + voxel_size * syn[4]) * 1e3

                        if synapse_obj:
                            obj = synapse_obj.copy()
                            if synapse_obj.data:
                                obj.data = synapse_obj.data.copy()
                            obj.animation_data_clear()
                            obj.location = (x, y, z)
                            obj.name = f"synapse-{n_synapses}"
                            VisualiseNetworkOld.link_object(obj)

                        else:
                            bpy.ops.mesh.primitive_uv_sphere_add(location=(x, y, z), size=0.001 * 4)
                            obj = bpy.context.selected_objects[0]
                            obj.active_material = mat_synapse
                            obj.select = False
                            synapse_obj = obj

                        n_synapses += 1

                        # print(f"Added synapse #{n_synapses} at {[x, y, z]}")
                        if n_synapses % 5000 == 0:
                            print(f"Synapses added so far: {n_synapses}")

            print(f"nSynapses = {n_synapses}")

        # Add a light source

        lamp_data = bpy.data.lamps.new(name="Sun", type='SUN')
        lamp_object = bpy.data.objects.new(name="Sun", object_data=lamp_data)
        bpy.context.scene.objects.link(lamp_object)

        # Place lamp to a specified location
        lamp_object.location = (1000.0, 1000.0, 1000.0)

        # Reposition camera
        cam = bpy.data.objects["Camera"]
        # cam.location = (2.98,2.68,4.96)
        # cam.rotation_euler = (1.59,0,-0.26)

        cam.location = (3.19, 3.46, 4.96)
        cam.rotation_euler = (96.7 * np.pi / 180, 0, -14.3 * np.pi / 180)

        # Is this needed?
        bpy.context.scene.update()

        bpy.ops.wm.save_as_mainfile(filepath=self.blender_save_file)

        if self.blender_output_image:
            print("Rendering image.")
            bpy.ops.render.render()
            bpy.data.images['Render Result'].save_render(filepath=self.blender_output_image)

    @staticmethod
    def copy_children(parent, parent_copy):
        for child in parent.children:
            child_copy = child.copy()
            child_copy.parent = parent_copy
            VisualiseNetworkOld.link_object(child_copy)
            VisualiseNetworkOld.copy_children(child, child_copy)

    @staticmethod
    def link_object(obj):
        try:
            bpy.context.scene.objects.link(obj)  # Blender 2.7
        except:
            print("Blender 2.7 failed, switch over to only use Blender 2.8 syntax")
            bpy.context.collection.objects.link(obj)  # Blender 2.8

    @staticmethod
    def clean_scene():
        # TODO: This does not seem to remove everything. Had some leftover synapses present still in notebook.
        print("Cleaning the scene.")
        del_list = bpy.context.copy()
        del_list['selected_objects'] = list(bpy.context.scene.objects)
        bpy.ops.object.delete(del_list)


import bpy
import mathutils
import numpy as np
import h5py

from snudda.utils.snudda_path import snudda_parse_path

# The script loads this position file and renders it
# Note that you might need to reposition the camera manually
# if your neurons are centred around a different position

# posFile = "../SmallTest2/network-neuron-positions.hdf5"
#posFile = "/home/hjorth/HBP/StriatumNetwork/model/Article-cube-2160/network-neuron-positions.hdf5"
posFile = "/home/hjorth/HBP/StriatumNetwork/model/Article2019HyperVoxel/network-neuron-positions.hdf5"

fi = h5py.File(posFile,"r")

neuronID = fi["network/neurons/neuronID"].value
pos = fi["network/neurons/position"].value
rot = fi["network/neurons/rotation"].value
morph = fi["network/neurons/morphology"]
name = fi["network/neurons/name"]

# Remove the start cube
bpy.ops.object.delete() 

bpy.data.scenes['Scene'].render.engine = 'CYCLES'
world = bpy.data.worlds['World']
world.use_nodes = True

# changing these values does affect the render.
bg = world.node_tree.nodes['Background']
bg.inputs[0].default_value[:3] = (1.0, 1.0, 1.0)
bg.inputs[1].default_value = 1.0

matMSD1 = bpy.data.materials.new("PKHG")
# matMSD1.diffuse_color = (1.0,0.2,0.2)
matMSD1.diffuse_color = (77./255,151./255,1.0)
matMSD2 = bpy.data.materials.new("PKHG")
# matMSD2.diffuse_color = (0.2,0.2,1.0)
matMSD2.diffuse_color = (67./255,55./255,181./255)

matFS = bpy.data.materials.new("PKHG")
matFS.diffuse_color = (6./255,31./255,85./255)
matChIN = bpy.data.materials.new("PKHG")
matChIN.diffuse_color = (252./255,102./255,0.0)
matLTS = bpy.data.materials.new("PKHG")
matLTS.diffuse_color = (150./255,63./255,212./255)

matOther = bpy.data.materials.new("PKHG")
matOther.diffuse_color = (0.4,0.4,0.4)

objLookup = dict([])

# Possible solution for extra speed:
# https://blender.stackexchange.com/questions/61094/copying-an-object-with-python-without-actually-creating-a-new-object
#

if(True):
  subset = np.random.permutation(2174)[:200]
  nTotal = len(subset)
else:
  subset = None
  nTotal = -1

ctr = 0


for ps,rt,mo,nm,nID in zip(pos,rot,morph,name,neuronID):

  if subset is not None and nID not in subset:
    continue

  ctr += 1
  
  if(mo in objLookup):
    print("Reusing obj for " + str(mo))
    obj = objLookup[mo].copy()
    #obj.data = objLookup[mo].data
    
    for o in objLookup[mo].children:
      oc = o.copy()
      # oc.data = o.data.copy()
      oc.parent = obj

      bpy.context.scene.objects.link(oc)
      
    bpy.context.scene.objects.link(obj)
    
  else:
    print("Loading morphology " + str(mo))
    bpy.ops.import_mesh.swc(filepath=snudda_parse_path(mo))
    obj = bpy.context.selected_objects[0]
    # obj = bpy.data.objects[-1]
    objLookup[mo] = obj

  eRot = mathutils.Matrix(rt.reshape(3,3)).to_euler()    
  obj.rotation_euler = eRot

  # Draw this neuron (the SWC import scales from micrometers to mm), the
  # positions in the simulation are in meters, need to scale it to mm for
  # blender to have same units.
  
  print("Setting position: " + str(ps*1e3))
  obj.location = ps*1e3

  nType = nm.decode().split("_")[0]
  if(nType == "dSPN"):
    print(str(nID) + " dSPN" + " - " + str(ctr) + "/" + str(nTotal))
    mat = matMSD1
  elif(nType == "iSPN"):
    print(str(nID) + " iSPN" + " - " + str(ctr) + "/" + str(nTotal))
    mat = matMSD2
  elif(nType == "FS"):
    print(str(nID) + " FS" + " - " + str(ctr) + "/" + str(nTotal))
    mat = matFS
  elif(nType == "ChIN"):
    print(str(nID) + " ChIN" + " - " + str(ctr) + "/" + str(nTotal))
    mat = matChIN
  elif(nType == "LTS"):
    print(str(nID) + " LTS" + " - " + str(ctr) + "/" + str(nTotal))
    mat = matLTS
  
  else:
    print(str(nID) + " other")
    mat = matOther


  for ch in obj.children:
    ch.active_material = mat

  obj.select = False
  
#import pdb
#pdb.set_trace()

# Add a light source

lampData = bpy.data.lamps.new(name="Sun", type='SUN')
lampObject = bpy.data.objects.new(name="Sun", object_data=lampData)
bpy.context.scene.objects.link(lampObject)

# Place lamp to a specified location
lampObject.location = (1000.0, 1000.0, 1000.0)

# Reposition camera
cam = bpy.data.objects["Camera"]
cam.location = (2.98,2.68,4.96)
cam.rotation_euler = (1.59,0,-0.26)

bpy.context.scene.update()


if(False):
  bpy.ops.render.render()
  bpy.data.images['Render Result'].save_render(filepath="cube-of-neurons-with-neurites.png")




import os
import numpy as np
from numba import jit
import matplotlib.pyplot as plt

import h5py
# from elephant.spike_train_correlation import spike_time_tiling_coefficient
from snudda.analyse.spike_time_tiling_coefficient import spike_time_tiling_coefficient
# from neo import SpikeTrain as NeoSpikeTrain
# import quantities as pq

from snudda.utils.load_network_simulation import SnuddaLoadSimulation
from snudda.utils.load import SnuddaLoad


class AnalyseSpikeTrains:

    def __init__(self, network_file=None, output_file=None, input_file=None, network_path=None):

        if network_path is None:
            if network_file is not None:
                network_path = os.path.dirname(network_file)
            elif output_file is not None:
                if "simulation" in output_file:
                    network_path = os.path.dirname(os.path.dirname(output_file))
            elif input_file is not None:
                network_path = os.path.dirname(input_file)

        if network_path is not None:
            if network_file is None:
                putative_network_file = os.path.join(network_path, "network-synapses.hdf5")
                if os.path.exists(putative_network_file):
                    network_file = putative_network_file

            if output_file is None:
                putative_output_file = os.path.join(network_path, "simulation", "output.hdf5")
                if os.path.exists(putative_output_file):
                    output_file = putative_output_file

            if input_file is None:
                putative_input_file = os.path.join(network_path, "input-spikes.hdf5")
                if os.path.exists(putative_input_file):
                    input_file = putative_input_file

        self.network_path = network_path
        self.network_file = network_file
        self.output_file = output_file
        self.input_file = input_file

        if output_file is not None:
            self.output_data = SnuddaLoadSimulation(network_simulation_output_file=self.output_file)
        else:
            self.output_data = None

        if self.network_file is not None:
            self.network_data = SnuddaLoad(network_file=self.network_file)
        else:
            self.network_data = None

        if self.input_file is not None:
            self.input_data = h5py.File(self.input_file, "r")

    def calculate_sttc(self, spike_train_a, spike_train_b, dt, start_time=0, end_time=None):

        if end_time is None:
            end_time = self.get_end_time()

        idx_a = np.where(np.logical_and(start_time <= spike_train_a, spike_train_a <= end_time))[0]
        idx_b = np.where(np.logical_and(start_time <= spike_train_b, spike_train_b <= end_time))[0]

        return spike_time_tiling_coefficient(spiketrain_i=spike_train_a[idx_a], spiketrain_j=spike_train_b[idx_b],
                                             dt=dt, end_time=end_time, start_time=start_time)

    def get_end_time(self):
        return np.max(self.output_data.get_time())

    def calculate_sttc_all_to_all(self, spike_trains, n_spikes, dt, start_time=0, end_time=None):
        n_spike_trains = len(n_spikes)
        assert spike_trains.shape[0] == n_spike_trains

        corr = []

        if end_time is None:
            end_time = self.get_end_time()

        pruned_spike_trains = []
        for st in spike_trains:
            idx = np.where(np.logical_and(start_time <= st, st <= end_time))[0]
            pruned_spike_trains.append(st[idx].flatten())

        for i in range(0, n_spike_trains):
            if i % 50 == 0:
                print(f'{i} / {n_spike_trains}')
            for j in range(i+1, n_spike_trains):
                corr.append(spike_time_tiling_coefficient(spiketrain_a=pruned_spike_trains[i],
                                                          spiketrain_b=pruned_spike_trains[j],
                                                          dt=dt, start_time=start_time, end_time=end_time))
        print(f'{i} / {n_spike_trains}')

        return np.array(corr)

    def calculate_sttc_one_to_all(self, spike_train, spike_trains, n_spikes, dt, start_time=0, end_time=None):
        n_spike_trains = len(n_spikes)
        assert spike_trains.shape[0] == n_spike_trains
        corr = []

        if end_time is None:
            end_time = self.get_end_time()

        st = spike_train.flatten()
        idx = np.where(np.logical_and(start_time <= st, st <= end_time))[0]
        spike_train_b = st[idx]

        for i in range(1, n_spike_trains):
            idx_i = np.where(np.logical_and(start_time <= spike_trains[i, :], spike_trains[i, :] <= end_time))[0]
            spike_train_i = spike_trains[i, idx_i].flatten()

            corr.append(spike_time_tiling_coefficient(spiketrain_a=spike_train_i,
                                                      spiketrain_b=spike_train_b,
                                                      dt=dt, end_time=end_time))

        return np.array(corr)

    def input_correlation(self, neuron_id, input_type, dt):
        input_spikes = self.input_data[f"input/{neuron_id}/{input_type}/spikes"][()].copy()
        n_spikes = self.input_data[f"input/{neuron_id}/{input_type}/spikes"].attrs["num_spikes"].copy()

        corr = self.calculate_sttc_all_to_all(spike_trains=input_spikes, n_spikes=n_spikes, dt=dt)
        return corr

    def input_output_correlation(self, neuron_id, dt, start_time=0, end_time=None):

        input_data = dict()
        corr = dict()

        # First gather all input spikes
        for input_type, input_spikes in self.input_data[f"input/{neuron_id}"].items():
            input_data[input_type] = input_spikes["spikes"][()].copy(), input_spikes["spikes"].attrs["num_spikes"].copy()

        output_data = self.output_data.get_spikes(neuron_id=neuron_id)

        for input_type, (input_spikes, n_spikes) in input_data.items():
            corr[input_type] = self.calculate_sttc_one_to_all(spike_train=output_data,
                                                              spike_trains=input_spikes,
                                                              n_spikes=n_spikes,
                                                              dt=dt, start_time=start_time, end_time=end_time)
        return corr

    def plot_spike_multiplicity(self, neuron_id, input_type, jitter=0, start_time=None, end_time=None):

        mult_list, mult = self.calculate_spike_multiplicity(neuron_id=neuron_id, input_type=input_type, jitter=jitter,
                                                            start_time=None, end_time=None)

        plt.figure()
        plt.title(f"{self.network_path} {neuron_id} {input_type}")
        # plt.stairs(edges=np.arange(0, len(mult)+1), values=mult)
        plt.hist(mult_list, bins=50)
        plt.yscale('log')

    def calculate_spike_multiplicity(self, neuron_id, input_type, jitter=0, start_time=None, end_time=None):

        input_spikes = self.input_data[f"input/{neuron_id}/{input_type}/spikes"][()].flatten()
        return self.calculate_multiplicity_helper(input_spikes=input_spikes, jitter=jitter,
                                                  start_time=start_time, end_time=end_time)

    @staticmethod
    # @jit(nopython=True, fastmath=True, cache=True)
    def calculate_multiplicity_helper(input_spikes, jitter=0, start_time=None, end_time=None):

        assert (start_time is None) ^ (end_time is None) == False, "Either both start_time and end_time is set, or None"

        max_mult = input_spikes.shape[0]
        multiplicity = np.zeros((max_mult+1, ), dtype=int)

        input_spikes = input_spikes.flatten()
        if end_time is None:
            input_spikes = np.sort(input_spikes[np.where(input_spikes >= 0)[0]])
        else:
            input_spikes = np.sort(input_spikes[np.where(np.logical_and(start_time <= input_spikes,
                                                                        input_spikes <= end_time))[0]])

        mul_ctr = 1
        first_spike = input_spikes[0]
        max_mult = 1

        mult_list = []

        for idx in range(1, len(input_spikes)):
            if input_spikes[idx] <= first_spike + jitter:
                mul_ctr += 1
            else:
                mult_list.append(mul_ctr)
                multiplicity[mul_ctr] += 1
                max_mult = max(max_mult, mul_ctr)
                mul_ctr = 1
                first_spike = input_spikes[idx]

        # Need to add the last spike also
        mult_list.append(mul_ctr)
        multiplicity[mul_ctr] += 1
        max_mult = max(max_mult, mul_ctr)

        return mult_list, multiplicity[:max_mult+1]





# import bpy

import h5py
import numpy as np

# Start by loading the positions from file

fi = h5py.File("../FullStriatumPos/network-neuron-positions.hdf5","r")

xm = 0.005
ym = 0.0035
zm = 0.0027
wm = 0.0003

keepIdx = []

pos = fi["network/neurons/position"]
nColour = []

for ir,row in enumerate(pos):
  if(ir % 50000 == 0):
    print(str(ir) + "/" + str(pos.shape[0]))
  if(xm <= row[0] and row[0] <= xm + wm \
     and ym <= row[1] and row[1] <= ym + wm \
     and zm <= row[2] and row[2] <= zm + wm):
    keepIdx.append(ir)

    nType = fi["network/neurons/name"][ir].decode().split("_")[0]
    if(nType == "MSD1"):
      col = [1.0, 0.2, 0.2]
    elif(nType == "MSD2"):
      col = [0.2, 0.2, 1.0]
    elif(nType == "LTS"):
      col = [0.7, 0.0, 0.7]
    elif(nType == "FS"):
      col = [0.1, 0.1, 0.1]
    elif(nType == "ChIN"):
      col = [0.9, 0.7, 0.1]      
    else:
      col = [0.4, 0.4, 0.4]
      
    nColour.append(col)
    
insideCoords = pos[keepIdx,:]
neuronColours = np.array(nColour)
posCol = np.concatenate((insideCoords,neuronColours),axis=1)
np.savetxt("hypervoxelcoords2.txt",posCol,delimiter=",")

print("Next render... drawHyperCubeSomasRender.py")

import pdb
pdb.set_trace()
      



# This script uses the functions defined in Network_analyse.py
#
# Performs analysis of the Striatum
#
#

import os
import sys

import matplotlib.pyplot as plt
import numpy as np

from snudda.analyse.analyse import SnuddaAnalyse


class SnuddaAnalyseStriatum(SnuddaAnalyse):

    def __init__(self, sim_dir, volume_type="cube", side_len=300e-6):

        assert os.path.exists(sim_dir), f"SnuddaAnalyseStriatum: Error {sim_dir} does not exist!"

        if os.path.isfile(sim_dir):
            # We allow the user to also send in a hdf5 file as sim_dir...
            hdf5_file = sim_dir
            self.sim_dir = os.path.dirname(sim_dir)
        else:
            assert os.path.isdir(sim_dir), f"SnuddaAnalyseStriatum: Expected sim_dir {sim_dir} to be a directory"
            self.sim_dir = sim_dir
            hdf5_file = os.path.join(sim_dir, "network-synapses.hdf5")

        print(f"Loading {hdf5_file}")

        super().__init__(hdf5_file=hdf5_file, load_cache=True,
                         volume_type=volume_type,
                         side_len=side_len)

    ############################################################################

    # Validation. How well do the synapse location for dSPN and iSPN match
    # the experimental data from Straub,..., Sabatini 2016

    def plot_fslt_scum_dist(self, plot_fs=True, plot_lts=True):

        pair_list_list = [[("FS", "dSPN"), ("LTS", "dSPN")],
                          [("FS", "iSPN"), ("LTS", "iSPN")]]
        figure_name_list = ["synapseCumulativeDistance-FS-and-LTS-to-dSPN.png",
                            "synapseCumulativeDistance-FS-and-LTS-to-iSPN.png"]
        figure_colour_list = [(6. / 255, 31. / 255, 85. / 255),
                              (150. / 255, 63. / 255, 212. / 255)]
        fill_range = [[0, 100e-6], [50e-6, 250e-6]]

        plot_flag = (plot_fs, plot_lts)

        assert plot_fs or plot_lts, "You must plot either FS or LTS, or both"

        if not plot_fs:
            figure_name_list = [x.replace("FS-and-", "") for x in figure_name_list]
        if not plot_lts:
            figure_name_list = [x.replace("and-LTS-", "") for x in figure_name_list]

        for pairList, figName \
                in zip(pair_list_list, figure_name_list):

            plt.rcParams.update({'font.size': 22})
            fig = plt.figure()
            ax = plt.subplot(111)
            # fig.tight_layout()
            fig.subplots_adjust(bottom=0.15, left=0.15)

            for pair, figCol, fillR, plotMeFlag \
                    in zip(pairList, figure_colour_list, fill_range, plot_flag):

                if not plotMeFlag:
                    continue

                try:
                    pair_id = tuple([self.all_types.index(x) for x in pair])
                except:
                    import traceback
                    tstr = traceback.format_exc()
                    print(tstr)
                    continue

                cum_dist = np.cumsum(self.dend_position_bin[pair_id]) \
                           / np.sum(self.dend_position_bin[pair_id])

                # Don't plot the full range
                end_idx = np.where(self.dend_position_edges <= 400e-6)[0][-1]

                ax.plot(self.dend_position_edges[:end_idx] * 1e6, cum_dist[:end_idx],
                        color=figCol, label=pair[0], linewidth=3)

                fill_idx = np.where(np.logical_and(fillR[0] <= self.dend_position_edges,
                                                   self.dend_position_edges <= fillR[1]))[0]
                fill_start = fill_idx[0]
                fill_end = fill_idx[-1]

                # Add the area marking
                ax.fill_between(self.dend_position_edges[fill_idx] * 1e6,
                                np.zeros((len(fill_idx),)),
                                cum_dist[fill_idx], alpha=0.95, color=figCol,
                                label=None)

                ax.set_xlabel('Distance from soma ($\mu$m)')
                ax.set_ylabel('Cumulative distrib.')

                if plot_fs and plot_lts:
                    ax.set_title("Synapse locations onto " + pair[1])
                else:
                    ax.set_title("Synapses " + pair[0] + " to " + pair[1])

            if plot_fs and plot_lts:
                # Only do legend if both are in figure
                ax.legend(loc="lower right")

            plt.ion()
            plt.show()
            plt.draw()
            plt.pause(0.0001)

            self.save_figure(plt, figName)

    ############################################################################


def cli(sim_dir):

    nas = SnuddaAnalyseStriatum(sim_dir, volume_type="cube")

    # import pdb
    # pdb.set_trace()
    #
    # nas.plotNeurons(0,showSynapses=True)

    plotHenrike = True
    plotChIN = True
    plotLTS = True

    dist3D = False
    # dist3D = True

    # import pdb
    # pdb.set_trace()

    if os.path.isfile(sim_dir):
        network_path = os.path.dirname(sim_dir)
    else:
        network_path = sim_dir

    dump_connection_data_to_file = os.path.join(network_path, "connection_data.json")

    nas.plot_synapse_cum_dist_summary(pair_list=[("dSPN", "ChIN"),
                                                 ("iSPN", "ChIN"),
                                                 ("LTS", "ChIN")])

    nas.plot_synapse_cum_dist_summary(pair_list=[("dSPN", "dSPN"),
                                                 ("iSPN", "dSPN"),
                                                 ("FS", "dSPN"),
                                                 ("LTS", "dSPN"),
                                                 ("ChIN", "dSPN")])

    nas.plot_synapse_cum_dist_summary(pair_list=[("dSPN", "iSPN"),
                                                 ("iSPN", "iSPN"),
                                                 ("FS", "iSPN"),
                                                 ("LTS", "iSPN"),
                                                 ("ChIN", "iSPN")])

    if plotHenrike:
        y_max_H = None  # 0.5

        nas.plot_connection_probability("dSPN", "iSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[50e-6, 100e-6],
                                        exp_data=[3 / 47.0, 3 / 66.0],
                                        exp_data_detailed=[(3, 47), (3, 66)],
                                        y_max=y_max_H,
                                        dump_data_to_file=dump_connection_data_to_file)
        nas.plot_connection_probability("dSPN", "dSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[50e-6, 100e-6],
                                        exp_data=[5 / 19.0, 3 / 43.0],
                                        exp_data_detailed=[(5, 19), (3, 43)],
                                        y_max=y_max_H,
                                        dump_data_to_file=dump_connection_data_to_file)
        nas.plot_connection_probability("iSPN", "dSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[50e-6, 100e-6],
                                        exp_data=[13 / 47.0, 10 / 80.0],
                                        exp_data_detailed=[(13, 47), (10, 80)],
                                        y_max=y_max_H,
                                        dump_data_to_file=dump_connection_data_to_file)
        nas.plot_connection_probability("iSPN", "iSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[50e-6, 100e-6],
                                        exp_data=[14 / 39.0, 7 / 31.0],
                                        exp_data_detailed=[(14, 39), (7, 31)],
                                        y_max=y_max_H,
                                        dump_data_to_file=dump_connection_data_to_file)

    nas.plot_num_synapses_per_pair("dSPN", "dSPN")
    nas.plot_num_synapses_per_pair("dSPN", "iSPN")
    nas.plot_num_synapses_per_pair("iSPN", "dSPN")
    nas.plot_num_synapses_per_pair("iSPN", "iSPN")

    # !!! Check edge effects

    nas.plot_incoming_connections(neuron_type="iSPN", pre_type="FS")
    nas.plot_incoming_connections(neuron_type="iSPN", pre_type="ChIN")
    nas.plot_incoming_connections(neuron_type="iSPN", pre_type="LTS")

    nas.plot_incoming_connections(neuron_type="dSPN", pre_type="dSPN")
    nas.plot_incoming_connections(neuron_type="dSPN", pre_type="iSPN")
    nas.plot_incoming_connections(neuron_type="iSPN", pre_type="dSPN")
    nas.plot_incoming_connections(neuron_type="iSPN", pre_type="iSPN")

    if True:
        # 2-5 ChIN should connect to each MS (approx)
        nas.plot_incoming_connections(neuron_type="dSPN", pre_type="ChIN")
        nas.plot_incoming_connections(neuron_type="iSPN", pre_type="ChIN")

    if True:
        nas.plot_connection_probability("FS", "iSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[100e-6, 150e-6, 250e-6],
                                        exp_data=[6 / 9.0, 21 / 54.0, 27 / 77.0],
                                        exp_data_detailed=[(6, 9), (21, 54), (27, 77)],
                                        y_max=None,
                                        dump_data_to_file=dump_connection_data_to_file)

        nas.plot_connection_probability("FS", "dSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[100e-6, 150e-6, 250e-6],
                                        exp_data=[8 / 9.0, 29 / 48.0, 48 / 90.0],
                                        exp_data_detailed=[(8, 9), (29, 48), (48, 90)],
                                        y_max=None,
                                        dump_data_to_file=dump_connection_data_to_file)

        nas.plot_num_synapses_per_pair("FS", "dSPN")
        nas.plot_num_synapses_per_pair("FS", "iSPN")

        #  Gittis,...,Kreitzer 2010 (p2228) -- 7/12 (and 3/4 reciprocal) -- distance?
        # FS->FS synapses weaker, 1.1 +/- 1.5nS

        nas.plot_connection_probability("FS", "FS",
                                        dist_3d=dist3D,
                                        exp_max_dist=[250e-6],
                                        exp_data=[7 / 12.0],
                                        exp_data_detailed=[(7, 12)],
                                        dump_data_to_file=dump_connection_data_to_file)

        nas.plot_num_synapses_per_pair("FS", "FS")

        # Koos & Tepper 1999, 2/6
        # McKeon, ... , Matheur 2022, 6/78 -- coupling 0.0006 to 0.0789
        nas.plot_connection_probability("FS", "FS",
                                        dist_3d=dist3D,
                                        connection_type="gapjunctions",
                                        exp_max_dist=[200e-6, 250e-6, 250e-6],
                                        exp_data=[6/78.0, 2 / 6.0, 3 / 7.0],
                                        exp_data_detailed=[(6, 78), (2, 6), (3, 7)],
                                        dump_data_to_file=dump_connection_data_to_file)

        nas.plot_num_synapses_per_pair("FS", "FS", connection_type="gapjunctions")

        nas.plot_incoming_connections(neuron_type="FS", pre_type="FS",
                                      connection_type="gapjunctions")

    nas.plot_fslt_scum_dist()
    nas.plot_fslt_scum_dist(plot_fs=False)
    nas.plot_fslt_scum_dist(plot_lts=False)

    nas.plot_num_synapses_per_pair("dSPN", "ChIN")
    nas.plot_num_synapses_per_pair("iSPN", "ChIN")

    nas.plot_incoming_connections(neuron_type="FS", pre_type="FS")
    nas.plot_incoming_connections(neuron_type="FS", pre_type="FS")

    nas.plot_num_synapses_per_pair("ChIN", "FS")

    nas.plot_synapse_cum_dist()

    if plotLTS:
        # 3/21 LTS->MS, Basal Ganglia book --- distance??
        # Ibanez-Sandoval, ..., Tepper  2011 3/21 -- if patching around visual axon
        # but 2/60 when patching blind
        nas.plot_connection_probability("LTS", "dSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[250e-6],
                                        exp_data=[2 / 60.0],
                                        exp_data_detailed=[(2, 60)],
                                        x_max=500,
                                        dump_data_to_file=dump_connection_data_to_file)

        nas.plot_connection_probability("LTS", "iSPN",
                                        dist_3d=dist3D,
                                        exp_max_dist=[250e-6],
                                        exp_data=[2 / 60.0],
                                        exp_data_detailed=[(2, 60)],
                                        x_max=500,
                                        dump_data_to_file=dump_connection_data_to_file)

        # Silberberg et al 2013, 2/12 FS-> LTS connected --- distance??
        # Voltage deflection... 0.5mV and 0.8mV
        # (check Szydlowski et al 2013, what Cl rev)
        #
        nas.plot_connection_probability("FS", "LTS",
                                        dist_3d=dist3D,
                                        exp_max_dist=[250e-6],
                                        exp_data=[2.0 / 12],
                                        exp_data_detailed=[(2, 12)],
                                        dump_data_to_file=dump_connection_data_to_file)

        nas.plot_num_synapses_per_pair("LTS", "dSPN")
        nas.plot_num_synapses_per_pair("LTS", "iSPN")
        nas.plot_num_synapses_per_pair("LTS", "ChIN")
        nas.plot_num_synapses_per_pair("FS", "LTS")

        # This plots figures for the article

    # 100e-6 from Planert 2010, and 250e-6 data from Gittis 2010
    # 150e-6 from Gittis 2011 (actually 100 +/- 50 micrometers)

    # MS <-> MS

    # FS -> MS

    nas.plot_num_synapses_per_pair("ChIN", "dSPN")
    nas.plot_num_synapses_per_pair("ChIN", "iSPN")
    nas.plot_num_synapses_per_pair("ChIN", "LTS")

    nas.plot_connection_probability("ChIN", "LTS",
                                    dist_3d=dist3D,
                                    dump_data_to_file=dump_connection_data_to_file)

    # Janicova 2015?? --- distance??!
    nas.plot_connection_probability("ChIN", "iSPN",
                                    dist_3d=dist3D,
                                    exp_max_dist=[250e-6],
                                    exp_data=[0.05],
                                    dump_data_to_file=dump_connection_data_to_file)

    nas.plot_connection_probability("ChIN", "dSPN",
                                    dist_3d=dist3D,
                                    exp_max_dist=[250e-6],
                                    exp_data=[0.05],
                                    dump_data_to_file=dump_connection_data_to_file)

    if True:
        nas.plot_connection_probability("LTS", "ChIN",
                                        dist_3d=dist3D,
                                        dump_data_to_file=dump_connection_data_to_file)

        # ALSO ADD GAP JUNCTIONS PLOT!!!
        # No exp data for this -- Gittis,...,Kreitzer 2010 (p2228) -- 7/12 (and 3/4 reciprocal) -- distance?
        # FS->FS synapses weaker, 1.1 +/- 1.5nS

    if plotChIN:
        # "I Janickova et al. 2017 så har de 2018 varicosities i en area på 655 um²,
        # deras slices är 70 um tjocka och om man antar att det inte är några
        # varicositites som täcker varandra så är volym-densiteten/mm³: 4.4*10⁷/mm3"
        # 1.7e6/24*0.01 = 708 ChIN per mm3
        # 4.4e7 / 708 = 62000 varicosities per ChIN
        #
        # 325 ChIN synapser per MS
        # 2-5 ChIN per MS
        # --> 65-160 synapser between a ChIN-MS pair
        # --> Each ChIN connect to 400 - 950 MS
        #
        # Number of MS within 350 micrometer radius 4*pi*(350e-6)^3/3*1.76e6/24e-9
        # --> 13100 MS reachable by ChIN at most (or rather number of MS somas
        # within radius of axonal arbour)
        # -->  3-7% connectivity probability??

        # nas.plotConnectionProbability("ChIN","iSPN", \
        #                              dist3D=dist3D,
        #                              expMaxDist=[200e-6],
        #                              expData=[62/89.0],
        #                              expDataDetailed=[(62,89)],
        #                              yMax=1.0)
        # This is from a targeted experiment, when they looked at where axon were?
        #
        # nas.plotConnectionProbability("ChIN","dSPN", \
        #                              dist3D=dist3D,
        #                              expMaxDist=[200e-6],
        #                              expData=[62/89.0],
        #                              expDataDetailed=[(62,89)],
        #                              yMax=1.0)

        nas.plot_connection_probability("ChIN", "FS",
                                        dist_3d=dist3D,
                                        y_max=None,
                                        dump_data_to_file=dump_connection_data_to_file)

        # A MS neuron receives 1e4 assymetrical synapses (Kincaid et al 1998),
        # and 2500 symmetrical synapses (Ingham et al 1998). Symmetrical synapses
        # can be dopaminergic, cholinergic or GABAergic, with dopaminergic
        # being 13% (Roberts et al 2002). Assuming that cholinergic inputs are
        # a similar percentage, 650 symmetrical synapses per MS are not GABAergic.
        #
        # --> 0.13*2500 = 325 ChIN inputs to MS
        nas.plot_incoming_connections(neuron_type="ChIN", pre_type="dSPN")
        nas.plot_incoming_connections(neuron_type="ChIN", pre_type="iSPN")
        nas.plot_incoming_connections(neuron_type="ChIN", pre_type="LTS")

        # 2-5 ChIN should connect to each MS (approx) --- ref? ?!?!?!
        nas.plot_incoming_connections(neuron_type="dSPN", pre_type="ChIN")
        nas.plot_incoming_connections(neuron_type="iSPN", pre_type="ChIN")

        # Om vi antar 2000 MS skulle kunna nå varje ChIN, när 10% aktiverade
        # så är 200 MS aktiva, om 75% av ChIN känner av MS input
        # (1-p)^200 = 0.25 --> 0.7 %
        nas.plot_connection_probability("dSPN", "ChIN",
                                        dist_3d=dist3D,
                                        dump_data_to_file=dump_connection_data_to_file)
        nas.plot_connection_probability("iSPN", "ChIN",
                                        dist_3d=dist3D,
                                        dump_data_to_file=dump_connection_data_to_file)

        # nas.nearestPreNeighbourDistance("LTS","dSPN")
        # nas.nearestPreNeighbourDistance("LTS","iSPN")

    if True:
        nas.plot_incoming_connections(neuron_type="dSPN", pre_type="iSPN")
        nas.plot_incoming_connections(neuron_type="dSPN", pre_type="dSPN")
        nas.plot_incoming_connections(neuron_type="dSPN", pre_type="FS")

        nas.plot_incoming_connections(neuron_type="iSPN", pre_type="iSPN")
        nas.plot_incoming_connections(neuron_type="iSPN", pre_type="dSPN")
        nas.plot_incoming_connections(neuron_type="iSPN", pre_type="FS")

        nas.plot_incoming_connections(neuron_type="dSPN", pre_type="LTS")
        nas.plot_incoming_connections(neuron_type="iSPN", pre_type="LTS")
        nas.plot_incoming_connections(neuron_type="ChIN", pre_type="LTS")

        nas.plot_incoming_connections(neuron_type="LTS", pre_type="ChIN")
        nas.plot_incoming_connections(neuron_type="LTS", pre_type="FS")

        nas.plot_incoming_connections(neuron_type="ChIN", pre_type="dSPN")
        nas.plot_incoming_connections(neuron_type="ChIN", pre_type="iSPN")

    # nas.plot_synapse_dist(density_flag=True)

if __name__ == "__main__":
    if len(sys.argv) > 1:
        sim_dir = sys.argv[1]
        print(f"Reading network from {sim_dir}")
    else:
        print("Please specify which directory the striatum network files is in")
        sys.exit(-1)

    cli(sim_dir)



import os
import json
import collections
import numpy as np
import h5py

import timeit

from snudda.utils import SnuddaLoad
from snudda.detect import SnuddaPrune

from scipy.stats import binomtest
from scipy.spatial import distance_matrix
from scipy.optimize import differential_evolution
from snudda.utils.numpy_encoder import NumpyEncoder

import uuid


class OptimisePruning:

    """ Optimises pruning parameters. First it creates a relatively small network and does touch detection on it.
        After that it does multiple pruning attempts (while keeping the files original detection files). """

    def __init__(self, network_path, pop_size=10, epochs=10, verbose=False):

        self.network_path = network_path
        self.pop_size = pop_size
        self.epochs = epochs

        self.prune = SnuddaPrune(network_path=network_path, keep_files=True)

        self.merge_files_syn = None
        self.merge_neuron_range_syn = None
        self.merge_syn_ctr = None
        self.merge_files_gj = None
        self.merge_neuron_range_gj = None
        self.merge_gj_ctr = None

        self.optimisation_info = dict()

        self.log_file = None
        self.verbose = verbose

    def merge_putative_synapses(self, force_merge=False):

        if force_merge:
            merge_info = None
        else:
            merge_info = self.prune.get_merge_info()

        if merge_info:
            merge_files_syn, merge_neuron_range_syn, merge_syn_ctr, \
            merge_files_gj, merge_neuron_range_gj, merge_gj_ctr = merge_info
        else:
            # From the hyper voxels gather all synapses (and gap junctions) belonging to specific neurons
            merge_files_syn, merge_neuron_range_syn, merge_syn_ctr, \
            merge_files_gj, merge_neuron_range_gj, merge_gj_ctr = self.prune.gather_neuron_synapses()

            self.prune.save_merge_info(merge_files_syn=merge_files_syn,
                                       merge_neuron_range_syn=merge_neuron_range_syn,
                                       merge_syn_ctr=merge_syn_ctr,
                                       merge_files_gj=merge_files_gj,
                                       merge_neuron_range_gj=merge_neuron_range_gj,
                                       merge_gj_ctr=merge_gj_ctr)

        self.merge_files_syn = merge_files_syn
        self.merge_neuron_range_syn = merge_neuron_range_syn
        self.merge_syn_ctr = merge_syn_ctr
        self.merge_files_gj = merge_files_gj
        self.merge_neuron_range_gj = merge_neuron_range_gj
        self.merge_gj_ctr = merge_gj_ctr

    def prune_synapses(self, pre_type, post_type, con_type, pruning_parameters, output_file):

        """ Prunes the network keeping only synapses between pre_type and post_type, using pruning_parameters

        Args:
            pre_type (str) : presynaptic neuron type (e.g. "FS")
            post_type (str) : postsynaptic neuron type (e.g. "dSPN")
            con_type (str) : type of connection (e.g. "GABA")
            pruning_parameters (dict) :
        """

        self.prune.load_pruning_information()

        orig_connectivity_distributions = json.loads(self.prune.hist_file["meta/connectivity_distributions"][()],
                                                     object_pairs_hook=collections.OrderedDict)

        # Next overwrite the pruning parameters
        pre_type_id = self.prune.type_id_lookup[pre_type]
        post_type_id = self.prune.type_id_lookup[post_type]
        orig_key = f"{pre_type}$${post_type}"
        synapse_type_id = orig_connectivity_distributions[orig_key][con_type]["channel_model_id"]

        pruning = self.prune.complete_pruning_info(pruning_parameters)
        pruning_other = None

        self.prune.connectivity_distributions = dict([])
        self.prune.connectivity_distributions[pre_type_id, post_type_id, synapse_type_id] = (pruning, pruning_other)

        # Update the config file in self.prune so the data gets written to network_synapses

        assert len(self.merge_files_syn) == 1, f"merge_Files_syn should be a list with one file only"

        # print(f"Writing to {output_file} (*)")

        # Clear the out file
        self.prune.out_file = None

        out_file = h5py.File(output_file, "w")

        if self.merge_files_syn[0] is not None:
            with h5py.File(self.merge_files_syn[0], "r") as f_syn:
                # print(f"Writing file {output_file}")
                # We need to make sure that we keep the pruned data files separate

                num_syn, num_syn_kept = self.prune.prune_synapses(synapse_file=f_syn,
                                                                  output_filename=out_file,
                                                                  row_range=None,
                                                                  close_out_file=False,
                                                                  close_input_file=True,
                                                                  merge_data_type="synapses")
        else:
            num_syn, num_syn_kept = self.prune.prune_synapses(synapse_file=None,
                                                              output_filename=out_file,
                                                              row_range=None,
                                                              close_out_file=False,
                                                              close_input_file=True,
                                                              merge_data_type="synapses")

        if self.merge_files_gj[0] is not None:
            with h5py.File(self.merge_files_gj[0], "r") as f_gj:

                num_gj, num_gj_kept = self.prune.prune_synapses(synapse_file=f_gj,
                                                                output_filename=out_file,
                                                                row_range=None,
                                                                close_out_file=False,
                                                                close_input_file=True,
                                                                merge_data_type="gap_junctions")
        else:
            num_gj, num_gj_kept = self.prune.prune_synapses(synapse_file=None,
                                                            output_filename=out_file,
                                                            row_range=None,
                                                            close_out_file=False,
                                                            close_input_file=True,
                                                            merge_data_type="gap_junctions")

        if not os.path.exists(output_file):
            print(f"Output file missing {output_file}")
            import pdb
            pdb.set_trace()

        try:
            assert output_file == self.prune.out_file.filename, f"Expected name {output_file}, had name {self.prune.out_file.name}"
        except:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()

        n_synapses = out_file["network/num_synapses"][()]
        n_gj = out_file["network/num_gap_junctions"][()]

        out_file["network/synapses"].resize((n_synapses, out_file["network/synapses"].shape[1]))
        out_file["network/gap_junctions"].resize((n_gj, out_file["network/gap_junctions"].shape[1]))

        out_file.close()

    def evaluate_fitness(self, pre_type, post_type, output_file, experimental_data,
                         mean_num_synapses_per_pair=None, std_num_synapses_per_pair=None,
                         percentile20_pair=None,
                         con_type="synapses", min_z=None, max_z=None):

        """

            Args:
                pre_type
                post_type
                output_file: path to output file from prune
                experiment_data: [(bin start, bin end, P)]
                mean_num_synapses_per_pair: avg_num

        """

        # print(f"Opening {output_file}")
        snudda_load = SnuddaLoad(network_file=output_file)
        snudda_data = snudda_load.data

        connection_matrix = np.zeros((snudda_data["num_neurons"], snudda_data["num_neurons"]), dtype=int)

        pre_id = snudda_load.get_neuron_id_of_type(neuron_type=pre_type)
        post_id = snudda_load.get_neuron_id_of_type(neuron_type=post_type)

        pre_mask = np.zeros((snudda_data["num_neurons"],), dtype=bool)
        post_mask = np.zeros((snudda_data["num_neurons"],), dtype=bool)

        pre_mask[pre_id] = True
        post_mask[post_id] = True

        # print(f"snudda_data.keys = {snudda_data.keys()}")

        n_synapses_cut_off = 0

        if con_type == "synapses":

            for row, coord in zip(snudda_data["synapses"], snudda_data["synapse_coords"]):
                if pre_mask[row[0]] and post_mask[row[1]]:
                    # TODO: Ignore synapses outside of cut-plane

                    if (min_z is not None and coord[2] < min_z) \
                        or (max_z is not None and coord[2] > max_z):

                        n_synapses_cut_off += 1
                        continue

                    # Only include connections between the right pre and post types
                    connection_matrix[row[0], row[1]] += 1

            n_kept = np.sum(connection_matrix)
            if self.verbose:
                print(f"{n_kept} synapses kept ({100.0*n_kept/(n_kept + n_synapses_cut_off):.2f} %), {n_synapses_cut_off} removed (outside slice)")

        else:
            assert min_z is None and max_z is None, "min_z and max_z is only currently supported for synapses, not gap junctions"

            for row in snudda_data[con_type]:
                if pre_mask[row[0]] and post_mask[row[1]]:
                    # TODO: Ignore synapses outside of cut-plane

                    # Only include connections between the right pre and post types
                    connection_matrix[row[0], row[1]] += 1

        pos = snudda_data["neuron_positions"]
        # We need to extract the parts relevant
        dist_matrix = distance_matrix(pos[pre_mask, :], pos[post_mask, :])

        n_connected = np.zeros((len(experimental_data),), dtype=int)
        n_total = np.zeros((len(experimental_data),), dtype=int)
        n_syn = 0
        n_pairs = 0
        n_syn_list = []

        # We also need to extract the relevant corresponding parts in connection_matrix
        con_mat = connection_matrix[pre_mask, :][:, post_mask]

        assert con_mat.shape == dist_matrix.shape, "Mismatch in shape"

        for dist, con in zip(dist_matrix.flatten(), con_mat.flatten()):

            if con > 0:
                n_syn += con
                n_pairs += 1
                n_syn_list.append(con)

            for idx, (bin_start, bin_end, _) in enumerate(experimental_data):
                if bin_start <= dist <= bin_end:
                    n_total[idx] += 1
                    if con > 0:
                        n_connected[idx] += 1

        n_syn_list = np.array(n_syn_list)

        # print(f"P={n_connected/n_total} .. {n_connected} .. {n_total}")

        p_hyp = np.zeros((len(experimental_data), ))

        for idx, (n_con, n_tot, p_exp) in enumerate(zip(n_connected, n_total, [x[2] for x in experimental_data])):
            # test = binomtest(n_con, n_tot, p_exp)
            # p_hyp[idx] = test.pvalue  # gave 0 when p values are too far apart, not informative
            p_hyp[idx] = abs(n_con/n_tot - p_exp) / (0.10 * p_exp)

        error = np.mean(p_hyp)

        per_pair_error = None
        std_pair_error = None
        percentile_error = None

        if mean_num_synapses_per_pair is not None:
            if n_pairs > 0:
                per_pair_error = np.abs(mean_num_synapses_per_pair - np.mean(n_syn_list)) / 2.0
            else:
                per_pair_error = abs(mean_num_synapses_per_pair)

            error +=  per_pair_error

        if std_num_synapses_per_pair is not None and n_pairs > 0:
            std_pair_error = np.abs(np.std(n_syn_list) - std_num_synapses_per_pair) / 0.5

            error += std_pair_error

        if percentile20_pair is not None and n_pairs > 0:
            percentile_error = np.abs(np.percentile(n_syn_list, 20) - percentile20_pair) / 0.1
            error += percentile_error

        if self.verbose:
            print(f"Errors in P: {np.mean(p_hyp)}, error in num con: {per_pair_error}, std_pair_error: {std_pair_error}, percentile20_error: {percentile_error}")

        return error

    @staticmethod
    def opt_helper_func(x, *args):

        pruning_parameters = dict()

        # args must be (optimisation_info)
        if args is None:
            raise ValueError("args must contain optimisation_info")
        else:
            optimisation_info = args[0]
            param_names = optimisation_info["param_names"]

            if "extra_pruning_parameters" in optimisation_info:
                pruning_parameters |= optimisation_info["extra_pruning_parameters"]

        for p_name, p_value in zip(param_names, x):
            pruning_parameters[p_name] = p_value

        if "output_file" in optimisation_info:
            output_file = optimisation_info["output_file"]
        else:
            output_file = os.path.join(optimisation_info["network_path"], "temp", f"network-synapses-{uuid.uuid4()}.hdf5")
        # print(f"Output file {output_file}")

        # Only include synapses between min_z and max_z
        min_z = optimisation_info.get("min_z", None)
        max_z = optimisation_info.get("max_z", None)

        # This trick allows us to reuse the same OptimisePruning object, will be faster
        op = OptimisePruning.get_op(optimisation_info)

        op.prune_synapses(pre_type=optimisation_info["pre_type"],
                          post_type=optimisation_info["post_type"],
                          con_type=optimisation_info["con_type"],
                          pruning_parameters=pruning_parameters,
                          output_file=output_file)

        if optimisation_info["con_type"].lower() in ("gap_junction", "gap_junctions"):
            con_type = "gap_junctions"
        else:
            con_type = "synapses"

        fitness = op.evaluate_fitness(pre_type=optimisation_info["pre_type"],
                                      post_type=optimisation_info["post_type"],
                                      output_file=output_file,
                                      experimental_data=optimisation_info["exp_data"],
                                      mean_num_synapses_per_pair=optimisation_info["mean_num_synapses_per_pair"],
                                      std_num_synapses_per_pair=optimisation_info["std_num_synapses_per_pair"],
                                      percentile20_pair=optimisation_info["percentile20_pair"],
                                      con_type=con_type,
                                      min_z=min_z, max_z=max_z)

        # print(f"Evaluating f1 = {x[0]}, fitness: {fitness}\n{output_file}\n")
        # print(f"Fitness: {fitness}")

        OptimisePruning.report_fitness(fitness)

        if "output_file" not in optimisation_info:
            os.remove(output_file)

        return fitness

    @staticmethod
    def report_fitness(fitness):

        OptimisePruning.ctr += 1
        OptimisePruning.fitness = min(OptimisePruning.fitness, fitness)

        if OptimisePruning.ctr % 100 == 0:
            print(f"Worker iter: {OptimisePruning.ctr}, fitness {OptimisePruning.fitness}")

    @staticmethod
    def get_op(optimisation_info):

        # I am sorry, this is ugly... but need to get this pickeable

        if "op" in vars(OptimisePruning) and OptimisePruning.op is not None:
            op = OptimisePruning.op
        else:
            op = OptimisePruning(network_path=optimisation_info["network_path"])
            merge_info = op.prune.get_merge_info()
            op.merge_files_syn, op.merge_neuron_range_syn, op.merge_syn_ctr, \
                op.merge_files_gj, op.merge_neuron_range_gj, op.merge_gj_ctr = merge_info

            OptimisePruning.op = op
            OptimisePruning.ctr = 0
            OptimisePruning.fitness = np.inf

        return op

    def optimize(self, pre_type, post_type, con_type,
                 experimental_data,
                 param_names, param_bounds,
                 extra_pruning_parameters,
                 avg_num_synapses_per_pair=None,
                 mean_num_synapses_per_pair=None,
                 std_num_synapses_per_pair=None,
                 percentile20_pair=None,
                 workers=1, maxiter=50, tol=0.001, pop_size=None,
                 min_z=None, max_z=None):

        start = timeit.default_timer()

        self.log_file = open(os.path.join(self.network_path, f"{pre_type}-{post_type}-{con_type}-optimisation-log.txt"), "wt")

        if pop_size is None:
            pop_size = self.pop_size

        if param_bounds == "default":
            param_bounds = []
            for p_name in param_names:
                if p_name == "softMax":
                    raise ValueError("Please change softMax to soft_max")

                if p_name == "f1":
                    param_bounds.append((0, 1))
                elif p_name == "soft_max":
                    param_bounds.append((0, 10))
                elif p_name == "mu2":
                    param_bounds.append((0, 5))
                elif p_name == "a3":
                    param_bounds.append((0, 1))
                else:
                    raise ValueError(f"No default parameter bounds for {p_name} (f1, softMax, mu2, a3)")

        # Let's warn user for a bit, this will change in future to mean_num_synapser_per_pair
        if avg_num_synapses_per_pair and mean_num_synapses_per_pair is None:
            print(f"OptimizePruning.optimize: avg_num_synapses_per_pair renamed to mean_num_synapses_per_pair, please change.")
            mean_num_synapses_per_pair = avg_num_synapses_per_pair

        self.optimisation_info["pre_type"] = pre_type
        self.optimisation_info["post_type"] = post_type
        self.optimisation_info["con_type"] = con_type
        self.optimisation_info["exp_data"] = experimental_data
        self.optimisation_info["mean_num_synapses_per_pair"] = mean_num_synapses_per_pair
        self.optimisation_info["std_num_synapses_per_pair"] = std_num_synapses_per_pair
        self.optimisation_info["percentile20_pair"] = percentile20_pair
        self.optimisation_info["extra_pruning_parameters"] = extra_pruning_parameters
        self.optimisation_info["ctr"] = 0
        self.optimisation_info["network_path"] = self.network_path
        self.optimisation_info["param_names"] = param_names
        self.optimisation_info["param_bounds"] = param_bounds
        self.optimisation_info["min_z"] = min_z
        self.optimisation_info["max_z"] = max_z

        optimisation_info = self.optimisation_info

        res = differential_evolution(func=OptimisePruning.opt_helper_func, args=(optimisation_info,),
                                     bounds=optimisation_info["param_bounds"], workers=workers,
                                     maxiter=maxiter, tol=tol, popsize=pop_size)

        # Rerun the best parameters, and keep data as network-synapses.hdf5
        optimisation_info["output_file"] = os.path.join(self.network_path, "network-synapses.hdf5")
        OptimisePruning.opt_helper_func(res.x, optimisation_info)

        duration = timeit.default_timer() - start
        self.log_file.write(f"Duration: {duration} s\n")
        print(f"Duration: {duration} s")

        self.log_file.close()

        # Clear the cache of OptimisePruning, so people can do other networks in same program
        if "op" in vars(OptimisePruning):
            del OptimisePruning.op

        return res

    @staticmethod
    def get_parameters(res, optimisation_info, truncate_decimals=False):

        param_names = optimisation_info["param_names"]

        pruning_parameters = dict()
        if "extra_pruning_parameters" in optimisation_info:
            pruning_parameters |= optimisation_info["extra_pruning_parameters"]

        for p_name, p_value in zip(param_names, res.x):
            if truncate_decimals:
                p_value = np.around(p_value, decimals=truncate_decimals)
            pruning_parameters[p_name] = p_value

        return pruning_parameters

    def export_json(self, file_name, res, append=False, truncate_decimals=4):

        pre_type = self.optimisation_info["pre_type"]
        post_type = self.optimisation_info["post_type"]
        connection_type = self.optimisation_info["con_type"]

        param_names = self.optimisation_info["param_names"]

        if append and os.path.isfile(file_name):
            with open(file_name, "r") as f:
                config_data = json.load(f)
        else:
            config_data = {"connectivity": {}}

        n_params = len(res.x)

        if f"{pre_type},{post_type}" in config_data["connectivity"]:
            con_data = config_data["connectivity"][f"{pre_type},{post_type}"]
        else:
            con_data = dict()

        if connection_type not in con_data:
            con_data[connection_type] = dict()

        con_data[connection_type]["pruning"] = self.get_parameters(res=res, optimisation_info=self.optimisation_info,
                                                                   truncate_decimals=truncate_decimals)

        if "pruning_other" in con_data[connection_type]:
            del con_data[connection_type]["pruning_other"]

        config_data["connectivity"][f"{pre_type},{post_type}"] = con_data

        with open(file_name, "w") as f:
            json.dump(config_data, f, cls=NumpyEncoder, indent=4)

